import datetime
import os
import time
from concurrent.futures.thread import ThreadPoolExecutor
from contextlib import contextmanager
from typing import (
    Iterable,
    Dict,
    List,
    Any,
    Union,
    Optional,
    Iterator,
    IO,
    ContextManager,
    Sequence,
)

from kong.drivers import InvalidJobStatus
from kong.logger import logger
from kong.util import rmtree, chunks
from .driver_base import DriverBase, checked_job
from ..model import Job
from ..db import database


class BatchDriverBase(DriverBase):
    def bulk_create_jobs(self, jobs: Iterable[Dict[str, Any]]) -> List["Job"]:
        return [self.create_job(**kwargs) for kwargs in jobs]

    def sync_status(self, job: "Job") -> Job:
        return self.bulk_sync_status([job])[0]

    def bulk_kill(self, jobs: Sequence["Job"]) -> Sequence["Job"]:
        now = datetime.datetime.now()
        jobs = self.bulk_sync_status(jobs)

        def delete() -> Iterable["Job"]:
            for job in jobs:
                self.kill(job, save=False)
                job.updated_at = now
                yield job

        with database.atomic():
            Job.bulk_update(
                delete(),
                fields=[Job.status, Job.updated_at],
                batch_size=self.batch_size,
            )

        return jobs

    def wait_gen(
        self,
        job: Union["Job", List["Job"]],
        poll_interval: Optional[int] = None,
        timeout: Optional[int] = None,
    ) -> Iterable[List[Job]]:
        start = datetime.datetime.now()
        poll_interval = poll_interval or 30

        jobs: List[Job]
        if isinstance(job, Job):
            jobs = [job]
        elif isinstance(job, list):
            jobs = job
        else:
            raise TypeError("Argument is neither job nor list of jobs")

        # pre-check for status
        for job in jobs:
            if job.status == Job.Status.CREATED:
                raise ValueError(f"Job is in status {job.status}, cannot wait")

        logger.debug("Begin waiting for %d jobs", len(jobs))

        while True:
            now = datetime.datetime.now()
            delta: datetime.timedelta = now - start
            if timeout is not None:
                if delta.total_seconds() > timeout:
                    raise TimeoutError()

            logger.debug("Refreshing %d", len(jobs))
            jobs = list(self.bulk_sync_status(jobs))  # overwrite with updated
            # filter out all that are considered waitable
            remaining_jobs = [
                j
                for j in jobs
                if j.status
                not in (Job.Status.COMPLETED, Job.Status.FAILED, Job.Status.UNKNOWN)
            ]
            if len(remaining_jobs) == 0:
                logger.debug("Waiting completed")
                break
            yield jobs
            logger.debug(
                "Waiting. Elapsed time: %s, %d jobs remaining",
                delta,
                len(remaining_jobs),
            )
            time.sleep(poll_interval)

    def bulk_submit(self, jobs: Iterable["Job"]) -> None:
        now = datetime.datetime.now()

        def sub() -> Iterable[Job]:
            for job in jobs:
                assert job.driver == self.__class__, "Not valid for different driver"
                self.submit(job, save=False)
                job.updated_at = now
                yield job

        with database.atomic():
            Job.bulk_update(
                sub(),
                fields=[Job.batch_job_id, Job.status, Job.updated_at],
                batch_size=self.batch_size,
            )

    @checked_job  # type: ignore
    @contextmanager  # type: ignore
    def stdout(self, job: Job) -> Iterator[IO[str]]:
        with open(job.data["stdout"], "r") as fh:
            yield fh

    def stderr(self, job: "Job") -> ContextManager[None]:
        raise NotImplementedError("Stderr goes to stdout in slurm")

    def resubmit(self, job: "Job") -> "Job":
        logger.debug("Resubmit job %s", job)
        job = self.sync_status(job)
        if job.status not in (
            Job.Status.FAILED,
            Job.Status.COMPLETED,
            Job.Status.UNKNOWN,
        ):
            raise InvalidJobStatus(f"Job {job} not in valid status for resubmit")

        try:
            self.kill(job)  # attempt to kill
        except Exception:
            pass

        # need to make sure the output artifacts are gone, since we're reusing the same job dir
        for name in ["stdout"]:
            path = job.data[name]
            if os.path.exists(path):
                logger.debug("Removing %s", path)
                os.remove(path)
            assert not os.path.exists(path)

        for d in ["output_dir"]:
            path = job.data[d]
            if os.path.exists(path):
                logger.debug("Removing %s", path)
                rmtree(path)
                os.makedirs(path)

        # reset to created
        job.status = Job.Status.CREATED
        job.save()

        self.submit(job)  # this will reset the status
        return job

    def bulk_resubmit(
        self, jobs: Iterable["Job"], do_submit: bool = True
    ) -> Iterable["Job"]:

        logger.debug("Resubmitting jobs")

        jobs = self.bulk_sync_status(list(jobs))
        # check status is ok
        for job in jobs:
            if job.status not in (
                Job.Status.FAILED,
                Job.Status.COMPLETED,
                Job.Status.UNKNOWN,
            ):
                raise InvalidJobStatus(f"Job {job} not in valid status for resubmit")

        try:
            jobs = self.bulk_kill(jobs)  # attempt to kill
        except Exception:
            pass

        def clean(job: Job) -> Job:
            # need to make sure the output artifacts are gone, since we're reusing the same job dir
            for name in ["stdout"]:
                path = job.data[name]
                if os.path.exists(path):
                    logger.debug("Removing %s", path)
                    os.remove(path)
                assert not os.path.exists(path)

            for d in ["output_dir"]:
                path = job.data[d]
                if os.path.exists(path):
                    logger.debug("Removing %s", path)
                    rmtree(path)
                    os.makedirs(path)

            return job

        nthreads = 40
        logger.debug("Cleaning up on %d threads", nthreads)
        with ThreadPoolExecutor(nthreads) as ex:
            jobs = list(ex.map(clean, jobs))

        # update status
        with database.atomic():

            with database.atomic():
                now = datetime.datetime.now()

                def jobit() -> Iterable[Job]:
                    for job in jobs:
                        job.status = Job.Status.CREATED
                        job.updated_at = now
                        yield job

                Job.bulk_update(
                    jobit(),
                    fields=[Job.status, Job.updated_at],
                    batch_size=self.batch_size,
                )

        jobs = Job.bulk_select(
            Job.job_id, [j.job_id for j in jobs], batch_size=self.select_batch_size
        )
        if do_submit:
            self.bulk_submit(jobs)

        return jobs

    def cleanup(self, job: "Job") -> "Job":
        job = self.sync_status(job)
        if job.status in (Job.Status.SUBMITTED, Job.Status.RUNNING):
            raise InvalidJobStatus(f"Job {job} might be running, please kill first")

        logger.debug("Cleanup job %s", job)
        for d in ["log_dir", "output_dir"]:
            path = job.data[d]
            if os.path.exists(path):
                logger.debug("Path %s exists, attempting to delete", path)
                rmtree(path)
        return job

    def bulk_cleanup(self, jobs: Sequence["Job"]) -> List["Job"]:
        jobs = self.bulk_sync_status(jobs)
        # safety check
        for job in jobs:
            assert job.driver == self.__class__
            if job.status in (Job.Status.SUBMITTED, Job.Status.RUNNING):
                raise InvalidJobStatus(f"Job {job} might be running, please kill first")

        logger.debug("Cleaning up %d jobs", len(jobs))

        for job in jobs:
            for d in ["log_dir", "output_dir"]:
                try:
                    path = job.data[d]
                    if os.path.exists(path):
                        logger.debug("Path %s exists, attempting to delete", path)
                        rmtree(path)
                except Exception:
                    logger.error("Unable to remove directory %s", d)
        return list(jobs)

    def remove(self, job: "Job") -> None:
        logger.debug("Removing job %s", job)
        job = self.cleanup(job)
        job.delete_instance()

    def bulk_remove(self, jobs: Sequence["Job"]) -> None:
        logger.debug("Removing %d jobs", len(jobs))
        jobs = self.bulk_cleanup(jobs)
        ids = [j.job_id for j in jobs]
        with database.atomic():
            for chunk in chunks(ids, self.select_batch_size):
                Job.delete().where(  # type: ignore
                    Job.job_id << chunk  # type: ignore
                ).execute()



from typing import Dict, cast

import peewee as pw
import sqlite3
import json

if sqlite3.sqlite_version_info < (3, 9, 0):  # type: ignore

    class JSONField(pw.CharField):  # pragma: no cover
        def db_value(self, value: Dict) -> str:
            return json.dumps(value)

        def python_value(self, value: str) -> Dict:
            return cast(Dict, json.loads(value))


else:  # pragma: no cover
    # flake8: noqa
    from playhouse.sqlite_ext import JSONField  # type: ignore



import math
import re
import os
import shutil
import stat
from datetime import timedelta
from typing import Optional, Any, TypeVar, Iterable, Iterator, List
import sys
import contextlib
from collections import deque

from tqdm import tqdm  # type: ignore
from halo import Halo  # type: ignore

from .logger import logger

strip = re.compile(r"\x1b(\[.*?[@-~]|\].*?(\x07|\x1b\\))")


def strip_colors(string: str) -> str:
    return strip.sub("", string)


def ljust(string: str, width: int, fillchar: str = " ") -> str:
    length = len(strip_colors(string))
    return string + (width - length) * fillchar


def rjust(string: str, width: int, fillchar: str = " ") -> str:
    length = len(strip_colors(string))
    return (width - length) * fillchar + string


def make_executable(path: str) -> None:
    mode = os.stat(path).st_mode
    os.chmod(path, mode | stat.S_IEXEC)


def is_executable(path: str) -> bool:
    mode = os.stat(path).st_mode
    return (mode & stat.S_IEXEC) != 0


def rmtree(path: str) -> None:
    # we'll try using shutil, and fall back to 'rm' if that fails
    try:
        shutil.rmtree(path)
    except OSError as e:
        logger.warning("shutil.rmtree failed: %s", e)
        # @TODO: This is absolutely not portable, maybe fix
        os.system(f"rm -rf {path}")


def format_timedelta(delta: timedelta) -> str:
    if delta >= timedelta(hours=100):
        raise ValueError(f"{delta} is too large to format")

    days = delta.days
    hours, rem = divmod(delta.seconds, 3600)
    minutes, seconds = divmod(rem, 60)

    total_hours = days * 24 + hours
    return f"{total_hours:02d}:{minutes:02d}:{seconds:02d}"


timedelta_regex = re.compile(r"\d\d:\d\d:\d\d")


def parse_timedelta(string: str) -> timedelta:
    if not timedelta_regex.match(string):
        raise ValueError(f"{string} does not have the right format")
    hours, minutes, seconds = string.split(":", 3)
    return timedelta(hours=int(hours), minutes=int(minutes), seconds=int(seconds))


@contextlib.contextmanager
def Spinner(
    text: str, persist: bool = True, *args: Any, **kwargs: Any
) -> Iterator[Halo]:
    stream = kwargs.get("stream", sys.stdout)
    if "spinner" not in kwargs:
        kwargs["spinner"] = "line"
        kwargs["interval"] = 80
    if stream.isatty():
        spinner = Halo(text, *args, **kwargs)  # type: ignore
        spinner.start()
        try:
            yield spinner
            if persist:
                spinner.succeed()
        except:  # noqa: E722
            if persist:
                spinner.fail()
            raise
        finally:
            if not persist:
                spinner.stop()
    else:
        sys.stdout.write(text + "\n")
        yield


T = TypeVar("T")


def Progress(*args: Any, **kwargs: Any) -> Iterable[T]:
    return tqdm(*args, **kwargs)  # type: ignore


def shorten(string: str, length: int) -> str:
    if length <= 4:
        raise ValueError("Shortening to <= 4 does not make sense")
    if length >= len(string):
        return string
    leftover = length - 3

    left_length = max(1, math.floor(leftover / 2))

    left = string[:left_length]
    right = string[-(length - left_length - 3) :]

    return f"{left}...{right}"


def shorten_path(path: str, last_length: Optional[int] = None) -> str:
    parts = path.split("/")
    shortened = [s[:1] for s in parts[:-1]]
    basename = parts[-1]
    if last_length is not None:
        basename = shorten(basename, last_length)
    shortened.append(basename)
    return "/".join(shortened)


def chunks(l: List[T], n: int) -> Iterable[List[T]]:
    for i in range(0, len(l), n):
        yield l[i : i + n]


def exhaust(generator: Iterable[Any]) -> None:
    deque(generator, maxlen=0)



from typing import Any, TypeVar, Type, List, Optional, Iterable


class Field:
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...

class CharField(Field): ...
class IntegerField(Field): ...
class DateField(Field): ...
class AutoField(Field): ...
class ForeignKeyField(Field): ...
class TimestampField(Field): ...
class DateTimeField(Field): ...

T = TypeVar('T', bound='Model')

class Model:
    def __init__(self, **kwargs: Any) -> None: ...
    @classmethod
    def create(cls, **kwargs: Any)-> T: ...
    def save(self, *args: Any, **kwargs: Any) -> None: ...
    @classmethod
    def get_or_none(cls: Type[T], *args: Any, **kwargs: Any) -> Optional[T]: ...
    def is_null(self) -> bool: ...
    def delete_instance(self, **kwargs: Any) -> None: ...
    @classmethod
    def bulk_update(cls, model_list: Iterable[T], fields: List[Any], batch_size: int) -> None: ...
    @classmethod
    def select(cls) -> Any: ...
    @classmethod
    def update(cls, **kwargs: Any) -> Any: ...

class SqliteDatabase:
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...
    def init(self, *args: Any) -> None: ...
    def connect(self)->None: ...
    def create_tables(self, tables: List[Any])->None: ...

class DoesNotExist(BaseException): ...
class IntegrityError(BaseException): ...




import os
from typing import Optional, Any, Dict

import yaml
import click

from . import config
from .logger import logger
from . import drivers
from .drivers import get_driver


def setup(cfg: Optional[config.Config]) -> None:
    logger.debug("Running setup")
    if not os.path.exists(config.APP_DIR):
        os.makedirs(config.APP_DIR)

    data: Dict[str, Any]
    if cfg is None:
        data = config.config_schema.validate({})
    else:
        data = dict(cfg.data)

    data["default_driver"] = click.prompt(
        f"Which batch system driver shall we use by default?",
        default=data.get("default_driver"),
    )

    try:
        assert issubclass(
            get_driver(data["default_driver"]), drivers.driver_base.DriverBase
        ), "Please provide a valid driver"
    except Exception:
        raise ValueError(f"{data['default_driver']} is not a valid driver")

    data["jobdir"] = os.path.expanduser(
        click.prompt(
            "Where is a good place to put job (e.g. log) files?",
            default=data.get("jobdir"),
        )
    )

    if not os.path.exists(data["jobdir"]):
        os.makedirs(data["jobdir"])

    data["joboutputdir"] = os.path.expanduser(
        click.prompt(
            "Where is a good place to put job output files?",
            default=data.get("joboutputdir"),
        )
    )

    if not os.path.exists(data["joboutputdir"]):
        os.makedirs(data["joboutputdir"])

    data["history_length"] = data.get("history_length")

    data = config.config_schema.validate(data)

    logger.debug("Config: %s", data)
    with open(config.CONFIG_FILE, "w") as f:
        yaml.dump(data, f)



import os
import functools
import socket

import pytest
import click
from click.testing import CliRunner
from unittest.mock import Mock

from kong.db import database
from kong import model
from kong.model import Folder, Job
import kong.setup
import kong

skip_lxplus = pytest.mark.skipif(
    socket.gethostname().startswith("lxplus"),
    reason="Lxplus doesn't allow access to proc info",
)


@pytest.yield_fixture
def db():
    database.init(":memory:")
    database.connect()
    database.create_tables([getattr(model, m) for m in model.__all__])
    yield database
    database.close()


@pytest.fixture
def app_env(tmp_path, monkeypatch):
    app_dir = os.path.join(tmp_path, "app")
    config_path = os.path.join(app_dir, "config.yml")
    monkeypatch.setattr("kong.config.APP_DIR", app_dir)
    monkeypatch.setattr("kong.config.CONFIG_FILE", config_path)
    monkeypatch.setattr("kong.config.DB_FILE", os.path.join(app_dir, "database.sqlite"))
    monkeypatch.setattr("kong.repl.history_file", os.path.join(tmp_path, "histfile"))
    assert not os.path.exists(config_path)
    assert kong.config.APP_DIR == app_dir
    assert kong.config.CONFIG_FILE == config_path
    return app_dir, config_path, tmp_path


@pytest.fixture
def cli():
    """Yield a click.testing.CliRunner to invoke the CLI."""
    class_ = click.testing.CliRunner

    def invoke_wrapper(f):
        """Augment CliRunner.invoke to emit its output to stdout.

        This enables pytest to show the output in its logs on test
        failures.

        """

        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            echo = kwargs.pop("echo", False)
            result = f(*args, **kwargs)

            if echo is True:
                sys.stdout.write(result.output)

            # if result.exception is not None:
            # raise result.exception

            return result

        return wrapper

    class_.invoke = invoke_wrapper(class_.invoke)
    cli_runner = class_()

    yield cli_runner


@pytest.fixture
def tree(db, state):

    root = Folder.get_root()

    f1 = root.add_folder("f1")
    f2 = root.add_folder("f2")

    alpha = f2.add_folder("alpha")
    beta = f2.add_folder("beta")

    gamma = f2.add_folder("gamma")
    delta = gamma.add_folder("delta")
    f3 = root.add_folder("f3")
    omega = f3.add_folder("omega")
    return root


@pytest.fixture
def sample_jobs(tree, state):
    driver = state.default_driver
    jobs = []
    job = lambda f: jobs.append(
        driver.create_job(command="sleep 0.1", folder=Folder.find_by_path(state.cwd, f))
    )
    root = tree

    for i in range(3):
        job("/")

    for i in range(4):
        job("/f1")

    for i in range(4):
        job("/f2")

    for i in range(2):
        job("/f2/beta")

    return jobs


@pytest.fixture
def state(app_env, db, monkeypatch):
    app_dir, config_path, tmp_path = app_env
    with monkeypatch.context() as m:
        m.setattr(
            "click.prompt",
            Mock(
                side_effect=[
                    "kong.drivers.local_driver.LocalDriver",
                    os.path.join(app_dir, "joblog"),
                    os.path.join(app_dir, "joboutput"),
                ]
            ),
        )
        kong.setup.setup(None)
    cfg = kong.config.Config()
    _state = kong.state.State(cfg, Folder.get_root())
    return _state



import os
import shutil
from datetime import timedelta, datetime, date
from typing import Collection, Iterator
from unittest.mock import Mock, ANY, call

import pytest

from kong import util
from kong.config import Config, slurm_schema
from kong.drivers import InvalidJobStatus
from kong.drivers.slurm_driver import (
    SlurmInterface,
    SlurmDriver,
    SlurmAccountingItem,
    ShellSlurmInterface,
)
from kong.model import Job, Folder
from kong.util import is_executable, exhaust


@pytest.fixture
def driver(monkeypatch, state):
    # set some config values
    data = state.config.data.copy()
    data["slurm_driver"] = dict(
        account="pseudo_account", node_size=42, default_queue="somequeue"
    )
    state.config = Config(data)

    monkeypatch.setattr(SlurmInterface, "__abstractmethods__", set())
    monkeypatch.setattr(ShellSlurmInterface, "__init__", Mock(return_value=None))

    sif = ShellSlurmInterface()

    return SlurmDriver(state.config, sif)


def test_sacct_parse(driver, monkeypatch, state):
    sacct_output = """
5205197|FAILED|2:0
5205197.batch|FAILED|2:0
5205197.extern|COMPLETED|0:0
5205197.0|FAILED|2:0
5205206|FAILED|2:0
5205206.batch|FAILED|2:0
5205206.extern|COMPLETED|0:0
5205206.0|FAILED|2:0
5205209|FAILED|2:0
5205209.batch|FAILED|2:0
5205209.extern|COMPLETED|0:0
5205209.0|FAILED|2:0
5205223|FAILED|13:0
5205223.batch|FAILED|13:0
5205223.extern|COMPLETED|0:0
5205223.0|FAILED|13:0
5205350|FAILED|13:0
5205350.batch|FAILED|13:0
5205350.extern|COMPLETED|0:0
5205350.0|FAILED|13:0
5205355|PENDING|0:0
5205757|COMPLETED|0:0
5205757.batch|COMPLETED|0:0
5205757.extern|COMPLETED|0:0
5205757.0|COMPLETED|0:0
22822|NOCLUE|0:0
    """.strip()

    with monkeypatch.context() as m:

        mock = Mock(return_value=sacct_output.split("\n"))
        m.setattr(driver.slurm, "_sacct", mock)

        td = timedelta(days = 10)
        res = list(driver.slurm.sacct([], td))

        starttime = date.today() - td

        mock.assert_called_once_with(
            brief=True, noheader=True, parsable2=True, starttime=starttime, _iter=True
        )

        ref = [
            SlurmAccountingItem(5_205_197, Job.Status.FAILED, 2),
            SlurmAccountingItem(5_205_206, Job.Status.FAILED, 2),
            SlurmAccountingItem(5_205_209, Job.Status.FAILED, 2),
            SlurmAccountingItem(5_205_223, Job.Status.FAILED, 13),
            SlurmAccountingItem(5_205_350, Job.Status.FAILED, 13),
            SlurmAccountingItem(5_205_355, Job.Status.SUBMITTED, 0),
            SlurmAccountingItem(5_205_757, Job.Status.COMPLETED, 0),
            SlurmAccountingItem(22822, Job.Status.UNKNOWN, 0),
        ]

        assert len(ref) == len(res)
        for a, b in zip(ref, res):
            assert a == b

        batch_job_id = 5_207_375
        sbatch = Mock(return_value=f"Submitted batch job {batch_job_id}")
        m.setattr(driver.slurm, "_sbatch", sbatch)

        job = Job()
        job.data["batchfile"] = "somefile.sh"

        jid = driver.slurm.sbatch(job)

        sbatch.assert_called_once_with("somefile.sh")

        assert jid == batch_job_id
        job.batch_job_id = jid

        scancel = Mock()
        m.setattr(driver.slurm, "_scancel", scancel)
        driver.slurm.scancel(job)
        scancel.assert_called_once_with(batch_job_id)


def test_repr():
    sai = SlurmAccountingItem(1, Job.Status.UNKNOWN, 0)
    assert repr(sai) != ""


def test_create_job(driver, state):
    root = Folder.get_root()
    j1 = driver.create_job(
        command="sleep 1",
        folder=root,
        cores=1,
        name="job1",
        queue="somequeue",
        walltime=timedelta(hours=5),
    )
    assert j1.status == Job.Status.CREATED
    assert len(root.jobs) == 1 and root.jobs[0] == j1
    assert j1.batch_job_id is None
    assert os.path.exists(j1.data["log_dir"])
    assert os.path.exists(j1.data["output_dir"])
    assert os.path.exists(j1.data["jobscript"])
    assert os.path.exists(j1.data["batchfile"])
    assert is_executable(j1.data["jobscript"])

    j2 = driver.create_job(
        command="sleep 1", walltime="03:00:00", folder=root, licenses="bliblablubb"
    )
    assert j2.data["walltime"] == "03:00:00"

    with open(j2.data["jobscript"]) as f:
        jobscript = f.read()
        assert str(j2.job_id) in jobscript
        assert str(j2.cores) in jobscript
        assert j2.command in jobscript
        for v in ["output_dir", "log_dir", "stdout"]:
            assert j2.data[v] in jobscript
    with open(j2.data["batchfile"]) as f:
        batchfile = f.read()
        assert str(j2.cores) in batchfile
        assert str(j2.memory) in batchfile
        for v in [
            "name",
            "slurm_out",
            "queue",
            "ntasks",
            "nnodes",
            "walltime",
            "account",
            "jobscript",
            "licenses",
        ]:
            assert str(j2.data[v]) in batchfile

    with pytest.raises(ValueError):
        driver.create_job(command="sleep 1", walltime="100:00:00", folder=root)

    with pytest.raises(ValueError):
        driver.create_job(command="sleep 1", walltime=42, folder=root)


def test_submit_job(driver, state, monkeypatch):
    root = Folder.get_root()
    j1 = driver.create_job(
        command="sleep 1",
        folder=root,
        cores=1,
        name="job1",
        queue="somequeue",
        walltime=timedelta(hours=5),
    )

    assert j1.status == Job.Status.CREATED

    j1.status = Job.Status.SUBMITTED
    j1.save()

    with pytest.raises(InvalidJobStatus):
        driver.submit(j1)

    j1.status = Job.Status.CREATED
    j1.save()

    batch_job_id = 5_207_375
    with monkeypatch.context() as m:
        sbatch = Mock(return_value=f"Submitted batch job {batch_job_id}")
        m.setattr(driver.slurm, "_sbatch", sbatch)
        driver.submit(j1)
        sbatch.assert_called_once_with(j1.data["batchfile"])

    assert j1.status == Job.Status.SUBMITTED
    assert j1.batch_job_id == str(batch_job_id)


def test_resubmit_job(driver, state, monkeypatch):
    root = Folder.get_root()
    j1 = driver.create_job(command="sleep 1", folder=root)

    assert j1.status == Job.Status.CREATED

    batch_job_id = 5_207_375
    sbatch = Mock(return_value=batch_job_id)
    monkeypatch.setattr(driver.slurm, "sbatch", sbatch)
    driver.submit(j1)
    sbatch.assert_called_once_with(j1)

    assert j1.status == Job.Status.SUBMITTED
    assert j1.batch_job_id == str(batch_job_id)

    monkeypatch.setattr(driver.slurm, "sacct", Mock(return_value=[]))
    with pytest.raises(InvalidJobStatus):
        driver.resubmit(j1)

    SAI = SlurmAccountingItem
    monkeypatch.setattr(
        driver.slurm,
        "sacct",
        Mock(return_value=[SAI(j1.batch_job_id, Job.Status.FAILED, 0)]),
    )

    bjid2 = 42
    sbatch = Mock(return_value=bjid2)
    monkeypatch.setattr(driver.slurm, "sbatch", sbatch)

    with monkeypatch.context() as m:
        # job errors on kill, resubmits anyway
        m.setattr(driver, "kill", Mock(side_effect=RuntimeError()))
        m.setattr("os.path.exists", Mock(side_effect=[True, False, False]))
        m.setattr("os.remove", Mock())
        j1 = driver.resubmit(j1)

    sbatch.assert_called_once()
    assert j1.status == Job.Status.SUBMITTED
    assert j1.batch_job_id == str(bjid2)  # gets new batch job id

    with monkeypatch.context() as m:
        m.setattr(driver, "sync_status", Mock())  # disable sync for a second
        with pytest.raises(InvalidJobStatus):
            driver.resubmit(j1)  # stays in SUBMITTED, not accepted

    monkeypatch.setattr(
        driver.slurm,
        "sacct",
        Mock(return_value=[SAI(j1.batch_job_id, Job.Status.FAILED, 0)]),
    )

    # will go to failed

    bjid3 = 99
    sbatch = Mock(return_value=bjid3)
    monkeypatch.setattr(driver.slurm, "sbatch", sbatch)
    j1 = driver.resubmit(j1)
    sbatch.assert_called_once()
    assert j1.status == Job.Status.SUBMITTED
    assert j1.batch_job_id == str(bjid3)


def test_job_bulk_resubmit(driver, state, monkeypatch):
    root = Folder.get_root()

    jobs = [
        driver.create_job(
            command="echo 'begin'; sleep 0.2 ; echo 'end' ; exit 1", folder=root
        ),
        driver.create_job(
            command="echo 'begin'; sleep 0.2 ; echo 'end' ; exit 1", folder=root
        ),
        driver.create_job(
            command="echo 'begin'; sleep 0.2 ; echo 'end' ; exit 1", folder=root
        ),
    ]

    other_job = driver.create_job(
        command="echo 'begin'; sleep 0.2 ; echo 'end' ; exit 1", folder=root
    )
    other_job.status = Job.Status.COMPLETED
    other_job.save()

    jobs[0].status = Job.Status.FAILED
    jobs[0].save()

    sbatch = Mock(side_effect=[1, 2, 3])
    monkeypatch.setattr(driver.slurm, "sbatch", sbatch)
    driver.bulk_submit(jobs[1:])
    assert sbatch.call_count == 2

    for job in jobs[1:]:
        job.status = Job.Status.COMPLETED

        with open(job.data["stdout"], "w") as f:
            f.write("hurz")

        job.save()

    shutil.rmtree(jobs[0].data["output_dir"])

    # we need to prevent driver from actually calling submit
    submit = Mock()
    remove = Mock(wraps=os.remove)
    makedirs = Mock()
    with monkeypatch.context() as m:
        m.setattr(driver, "submit", submit)
        m.setattr(driver.slurm, "sacct", Mock(return_value=[]))
        m.setattr(driver, "bulk_kill", Mock(side_effect=RuntimeError))
        m.setattr("os.remove", remove)
        m.setattr("os.makedirs", makedirs)
        driver.bulk_resubmit(jobs)
    assert submit.call_count == len(jobs)
    remove.assert_has_calls([call(j.data["stdout"]) for j in jobs[1:]], any_order=True)
    makedirs.assert_has_calls(
        [call(j.data["output_dir"]) for j in jobs[1:]], any_order=True
    )

    for job in jobs:
        job.reload()
        assert job.status == Job.Status.CREATED

    # bug: all jobs where reset to created. Check this is not the case anymore
    other_job.reload()
    assert other_job.status != Job.Status.CREATED


def test_resubmit_bulk_invalid_status(driver, state, monkeypatch):
    monkeypatch.setattr(driver, "sync_status", Mock())
    j1 = driver.create_job(command="sleep 1", folder=state.cwd)
    for status in (Job.Status.CREATED, Job.Status.SUBMITTED, Job.Status.RUNNING):
        j1.status = status
        j1.save()
        with pytest.raises(InvalidJobStatus):
            driver.bulk_resubmit([j1])


def test_job_bulk_resubmit_no_submit(driver, state, monkeypatch):
    root = Folder.get_root()

    jobs = [
        driver.create_job(
            command="echo 'begin'; sleep 0.2 ; echo 'end' ; exit 1", folder=root
        ),
        driver.create_job(
            command="echo 'begin'; sleep 0.2 ; echo 'end' ; exit 1", folder=root
        ),
        driver.create_job(
            command="echo 'begin'; sleep 0.2 ; echo 'end' ; exit 1", folder=root
        ),
    ]

    sbatch = Mock(side_effect=[1, 2, 3])
    monkeypatch.setattr(driver.slurm, "sbatch", sbatch)
    driver.bulk_submit(jobs)
    assert sbatch.call_count == 3

    for job in jobs:
        job.status = Job.Status.COMPLETED
        job.save()

    bulk_submit = Mock()
    with monkeypatch.context() as m:
        m.setattr(driver.slurm, "sacct", Mock(return_value=[]))
        m.setattr(driver, "bulk_submit", bulk_submit)
        driver.bulk_resubmit(jobs, do_submit=False)
    assert bulk_submit.call_count == 0


def test_stdout_stderr(driver, state, monkeypatch):
    root = Folder.get_root()
    j1 = driver.create_job(
        command="sleep 1",
        folder=root,
        cores=1,
        name="job1",
        queue="somequeue",
        walltime=timedelta(hours=5),
    )

    assert j1.status == Job.Status.CREATED

    batch_job_id = 5_207_375
    sbatch = Mock(return_value=batch_job_id)
    monkeypatch.setattr(driver.slurm, "sbatch", sbatch)
    driver.submit(j1)

    assert j1.status == Job.Status.SUBMITTED
    assert j1.batch_job_id == str(batch_job_id)

    stdout = "VALUE VALUE VALUE"

    with open(j1.data["stdout"], "w") as fh:
        fh.write(stdout)

    with driver.stdout(j1) as fh:
        assert stdout == fh.read()

    with pytest.raises(NotImplementedError):
        driver.stderr(j1)


def test_sync_status(driver, monkeypatch):
    root = Folder.get_root()
    j1 = driver.create_job(command="sleep 1", folder=root)

    with monkeypatch.context() as m:
        sbatch = Mock(return_value=1)
        m.setattr(driver.slurm, "sbatch", sbatch)

    assert j1.status == Job.Status.CREATED

    batch_job_id = 5_207_375
    monkeypatch.setattr(driver.slurm, "sbatch", Mock(return_value=batch_job_id))
    driver.submit(j1)
    assert j1.status == Job.Status.SUBMITTED
    assert j1.batch_job_id == str(batch_job_id)

    sacct_return = [
        [SlurmAccountingItem(batch_job_id, Job.Status.RUNNING, 0)],
        [SlurmAccountingItem(batch_job_id, Job.Status.FAILED, 0)],
    ]
    sacct = Mock(side_effect=sacct_return)
    monkeypatch.setattr(driver.slurm, "sacct", sacct)

    j1 = driver.sync_status(j1)
    assert j1.status == Job.Status.RUNNING
    j1 = driver.sync_status(j1)
    assert j1.status == Job.Status.FAILED


def test_bulk_create(driver, state):
    root = Folder.get_root()
    jobs = driver.bulk_create_jobs(
        [{"folder": root, "command": "sleep 1"} for i in range(10)]
    )
    assert len(jobs) == 10
    for job in jobs:
        assert job.status == Job.Status.CREATED


def test_bulk_submit(driver, state, monkeypatch):
    root = Folder.get_root()

    jobs = [
        driver.create_job(folder=root, command=f"sleep 0.1; echo 'JOB{i}'")
        for i in range(15)
    ]

    assert len(jobs) == 15
    for job in jobs:
        assert job.status == Job.Status.CREATED

    sbatch = Mock(
        side_effect=[f"Submitted batch job {i + 1}" for i in range(len(jobs))]
    )
    monkeypatch.setattr(driver.slurm, "_sbatch", sbatch)
    driver.bulk_submit(jobs)

    assert sbatch.call_count == len(jobs)

    for job in jobs:
        assert job.status == Job.Status.SUBMITTED
        assert str(job.job_id) == job.batch_job_id


def test_bulk_sync_status(driver, state, monkeypatch):

    root = Folder.get_root()

    jobs = [
        driver.create_job(folder=root, command=f"sleep 0.1; echo 'JOB{i}'")
        for i in range(15)
    ]

    assert len(jobs) == 15
    for job in jobs:
        assert job.status == Job.Status.CREATED

    sbatch = Mock(side_effect=[i + 1 for i in range(len(jobs))])
    monkeypatch.setattr(driver.slurm, "sbatch", sbatch)
    driver.bulk_submit(jobs)

    sacct_return = ["|".join([str(i + 1), "RUNNING", "0:0"]) for i in range(len(jobs))]
    sacct = Mock(return_value=sacct_return)
    # pretend they're all running now
    monkeypatch.setattr(driver.slurm, "_sacct", sacct)

    jobs = driver.bulk_sync_status(jobs)

    sacct.assert_called_once_with(
        jobs=",".join([j.batch_job_id for j in jobs]),
        brief=True,
        noheader=True,
        parsable2=True,
        starttime=ANY,
        _iter=True,
    )

    for job in jobs:
        assert job.status == Job.Status.RUNNING

    sacct_return = [
        "|".join([str(i + 1), "COMPLETED" if i < 6 else "FAILED", "0:0"])
        for i in range(len(jobs))
    ]

    sacct = Mock(return_value=sacct_return)
    monkeypatch.setattr(driver.slurm, "_sacct", sacct)

    jobs = driver.bulk_sync_status(jobs)
    sacct.assert_called_once_with(
        jobs=",".join([j.batch_job_id for j in jobs]),
        brief=True,
        noheader=True,
        parsable2=True,
        starttime=ANY,
        _iter=True,
    )

    for job in jobs[:6]:
        assert job.status == Job.Status.COMPLETED
    for job in jobs[6:]:
        assert job.status == Job.Status.FAILED


def test_bulk_sync_status_invalid_id(driver, state, monkeypatch):

    root = Folder.get_root()

    jobs = driver.bulk_create_jobs(
        [{"folder": root, "command": "sleep 1"} for i in range(10)]
    )

    sbatch = Mock(side_effect=[i + 1 for i in range(len(jobs))])
    monkeypatch.setattr(driver.slurm, "sbatch", sbatch)
    driver.bulk_submit(jobs)

    SAI = SlurmAccountingItem
    sacct_return = [SAI(i + 1, Job.Status.RUNNING, 0) for i in range(len(jobs))]
    sacct_return += [SAI(12_345_665, Job.Status.UNKNOWN, 0)]
    sacct = Mock(return_value=sacct_return)
    # pretend they're all running now
    monkeypatch.setattr(driver.slurm, "sacct", sacct)
    jobs = driver.bulk_sync_status(jobs)

    for job in jobs:
        assert job.status == Job.Status.RUNNING


def test_kill_job(driver, state, monkeypatch):
    root = Folder.get_root()
    j1 = driver.create_job(command="sleep 1", folder=root)

    driver.kill(j1)
    assert j1.status == Job.Status.FAILED

    j1.status = Job.Status.CREATED

    monkeypatch.setattr(driver.slurm, "sbatch", Mock(return_value=1))
    driver.submit(j1)

    assert j1.status == Job.Status.SUBMITTED

    scancel = Mock()
    monkeypatch.setattr(driver.slurm, "_scancel", scancel)
    driver.kill(j1)
    scancel.assert_called_once_with(j1.batch_job_id)

    assert j1.status == Job.Status.FAILED

    driver.kill(j1)
    assert j1.status == Job.Status.FAILED


def test_bulk_kill(driver, state, monkeypatch):
    root = Folder.get_root()

    jobs = [
        driver.create_job(folder=root, command=f"sleep 0.1; echo 'JOB{i}'")
        for i in range(15)
    ]

    for job in jobs:
        assert job.status == Job.Status.CREATED

    sbatch = Mock(side_effect=[i for i in range(len(jobs))])
    monkeypatch.setattr(driver.slurm, "sbatch", sbatch)
    driver.bulk_submit(jobs)
    assert sbatch.call_count == len(jobs)

    for job in jobs:
        assert job.status == Job.Status.SUBMITTED

    scancel = Mock()
    monkeypatch.setattr(driver.slurm, "scancel", scancel)
    monkeypatch.setattr(driver.slurm, "sacct", Mock(return_value=[]))

    jobs = driver.bulk_kill(jobs)

    assert scancel.call_count == len(jobs)

    for job in jobs:
        assert job.status == Job.Status.FAILED


def test_wait(driver, state, monkeypatch):
    root = Folder.get_root()

    class SlurmInterfaceDummy(SlurmInterface):
        def __init__(self):
            self.max_job_id = 1
            self.id_map = {}
            self.state_idx = 0
            self.jobs = []

        def sacct(self, jobs: Collection["Job"], start_delta: timedelta) -> Iterator[SlurmAccountingItem]:
            values = [
                [
                    SlurmAccountingItem(j.batch_job_id, Job.Status.RUNNING, 0)
                    for j in self.jobs
                ],
                [  # first half done
                    SlurmAccountingItem(j.batch_job_id, Job.Status.COMPLETED, 0)
                    for j in self.jobs[len(self.jobs) // 2 :]
                ],
                [  # all done
                    SlurmAccountingItem(j.batch_job_id, Job.Status.COMPLETED, 0)
                    for j in self.jobs
                ],
            ]

            v = values[self.state_idx]
            self.state_idx += 1
            return v

        def sbatch(self, job: Job) -> int:
            self.jobs.append(job)
            self.max_job_id += 1
            job.batch_job_id = self.max_job_id
            self.id_map[job] = self.max_job_id
            return self.max_job_id

        def scancel(self, job: Job) -> None:
            pass

    si = SlurmInterfaceDummy()
    monkeypatch.setattr(driver, "slurm", si)

    jobs = [
        driver.create_job(folder=root, command=f"sleep 0.1; echo 'JOB{i}'")
        for i in range(15)
    ]

    for job in jobs:
        assert job.status == Job.Status.CREATED

    driver.bulk_submit(jobs)

    for job in jobs:
        job.reload()
        i = si.id_map[job]
        assert job.batch_job_id == str(i)
        assert job.status == Job.Status.SUBMITTED

    with monkeypatch.context() as m:
        sacct = Mock(wraps=si.sacct)
        m.setattr(si, "sacct", sacct)
        driver.wait(jobs, poll_interval=0.01)
        assert sacct.call_count == 3

    for job in jobs:
        job.reload()
        assert job.status == Job.Status.COMPLETED


def test_wait_single(driver, monkeypatch):
    root = Folder.get_root()
    j1 = driver.create_job(folder=root, command=f"sleep 0.1; echo 'JOB'")

    monkeypatch.setattr(driver, "bulk_sync_status", Mock(return_value=[j1]))

    with pytest.raises(ValueError):
        driver.wait(j1)  # job is in CREATED

    j1.status = Job.Status.COMPLETED
    driver.wait(j1)

    # timeout
    j1.status = Job.Status.RUNNING
    with pytest.raises(TimeoutError):
        driver.wait(j1, timeout=0.05, poll_interval=0.1)

    with pytest.raises(TypeError):
        driver.wait("nope")


def test_config_schema():
    assert slurm_schema.is_valid(dict())
    assert slurm_schema.is_valid(
        dict(account="bla", node_size=80, default_queue="whatever")
    )
    assert not slurm_schema.is_valid(dict(account="bla", node_size="blub"))
    assert not slurm_schema.is_valid(dict(account=42))


def test_cleanup_driver(driver, state, monkeypatch):
    root = Folder.get_root()
    j1 = driver.create_job(command="sleep 1", folder=root)

    assert j1.status == Job.Status.CREATED
    assert os.path.exists(j1.data["log_dir"])
    assert os.path.exists(j1.data["output_dir"])

    # disable job updates
    monkeypatch.setattr(driver, "sync_status", Mock(side_effect=lambda j: j))

    j1.status = Job.Status.SUBMITTED
    with pytest.raises(InvalidJobStatus):
        driver.cleanup(j1)
    assert os.path.exists(j1.data["log_dir"])
    assert os.path.exists(j1.data["output_dir"])
    j1.status = Job.Status.RUNNING
    with pytest.raises(InvalidJobStatus):
        driver.cleanup(j1)
    assert os.path.exists(j1.data["log_dir"])
    assert os.path.exists(j1.data["output_dir"])

    j1.status = Job.Status.COMPLETED

    driver.cleanup(j1)
    assert not os.path.exists(j1.data["log_dir"])
    assert not os.path.exists(j1.data["output_dir"])


def test_cleanup_driver_already_deleted(driver, state, monkeypatch):
    root = Folder.get_root()
    j1 = driver.create_job(command="sleep 1", folder=root)

    assert j1.status == Job.Status.CREATED
    assert os.path.exists(j1.data["log_dir"])
    assert os.path.exists(j1.data["output_dir"])

    shutil.rmtree(j1.data["log_dir"])

    # disable job updates
    monkeypatch.setattr(driver, "sync_status", Mock(side_effect=lambda j: j))

    j1.status = Job.Status.COMPLETED

    assert not os.path.exists(j1.data["log_dir"])
    rmtree = Mock(wraps=util.rmtree)
    with monkeypatch.context() as m:
        m.setattr("kong.drivers.batch_driver_base.rmtree", rmtree)
        driver.cleanup(j1)
    rmtree.assert_has_calls([call(j1.data["output_dir"])])

    assert not os.path.exists(j1.data["log_dir"])
    assert not os.path.exists(j1.data["output_dir"])


def test_job_bulk_cleanup(driver, state, monkeypatch):
    jobs = [
        driver.create_job(command="sleep 1", folder=state.cwd),
        driver.create_job(command="sleep 1", folder=state.cwd),
        driver.create_job(command="sleep 1", folder=state.cwd),
    ]

    for job in jobs:
        assert os.path.exists(job.data["log_dir"])
        assert os.path.exists(job.data["output_dir"])

    jobs[0].status = Job.Status.RUNNING
    jobs[0].save()

    with pytest.raises(InvalidJobStatus):
        driver.bulk_cleanup(jobs)

    for job in jobs:
        assert os.path.exists(job.data["log_dir"])
        assert os.path.exists(job.data["output_dir"])

    shutil.rmtree(jobs[0].data["log_dir"])
    jobs[0].status = Job.Status.CREATED
    jobs[0].save()

    rmtree = Mock(side_effect=OSError)
    with monkeypatch.context() as m:
        m.setattr("kong.drivers.batch_driver_base.rmtree", rmtree)
        driver.bulk_cleanup(jobs)
    rmtree.assert_has_calls(
        [
            call(jobs[0].data["output_dir"]),
            call(jobs[1].data["log_dir"]),
            call(jobs[1].data["output_dir"]),
            call(jobs[2].data["log_dir"]),
            call(jobs[2].data["output_dir"]),
        ]
    )

    driver.bulk_cleanup(jobs)

    for job in jobs:
        assert not os.path.exists(job.data["log_dir"])
        assert not os.path.exists(job.data["output_dir"])


def test_remove_driver(driver, state, monkeypatch):
    root = Folder.get_root()
    j1 = driver.create_job(command="sleep 1", folder=root)

    # disable job updates
    monkeypatch.setattr(driver, "sync_status", Mock(side_effect=lambda j: j))
    cleanup = Mock(side_effect=lambda j: j)
    monkeypatch.setattr(driver, "cleanup", cleanup)

    j1.status = Job.Status.COMPLETED

    driver.remove(j1)

    cleanup.assert_called_once_with(j1)

    assert Job.get_or_none(job_id=j1.job_id) is None


def test_job_bulk_remove(driver, state):
    jobs = [
        driver.create_job(command="sleep 1", folder=state.cwd),
        driver.create_job(command="sleep 1", folder=state.cwd),
        driver.create_job(command="sleep 1", folder=state.cwd),
    ]
    for job in jobs:
        assert os.path.exists(job.data["log_dir"]), "Does not create job directory"
        assert os.path.exists(
            job.data["output_dir"]
        ), "Does not create output directory"

    driver.bulk_remove(jobs)

    for job in jobs:
        assert not os.path.exists(
            job.data["log_dir"]
        ), "Driver does not cleanup job directory"
        assert not os.path.exists(
            job.data["output_dir"]
        ), "Driver does not cleanup output directory"



import functools
import os
from typing import (
    List,
    Any,
    Union,
    Optional,
    ContextManager,
    Iterable,
    TYPE_CHECKING,
    Dict,
    Collection,
    Sequence,
)
from abc import abstractmethod, ABC

from kong.drivers import DriverMismatch
from kong.util import exhaust
from ..logger import logger
from ..config import Config

if TYPE_CHECKING:  # pragma: no cover
    from ..model import Job, Folder


def checked_job(f: Any) -> Any:
    @functools.wraps(f)
    def wrapper(self: Any, job: "Job", *args: Any, **kwargs: Any) -> Any:
        self._check_driver(job)
        return f(self, job, *args, **kwargs)

    return wrapper


class DriverBase(ABC):  # pragma: no-cover
    config: Config

    batch_size: int = 50
    select_batch_size: int = 500

    def __init__(self, config: Optional["Config"]) -> None:
        if config is None:
            logger.debug("Attempt to default-construct configuration object")
            self.config = Config()
        else:
            logger.debug("Taking explicit config")
            self.config = config

        logger.debug("Checking jobdir filesystem at %s", self.config.jobdir)
        assert os.path.exists(self.config.jobdir)

    @classmethod
    def _check_driver(cls, job: "Job") -> None:
        # check if we're the right driver for this
        if cls != job.driver:
            raise DriverMismatch(f"Job {job} is has driver {job.driver}, not {cls}")

    @abstractmethod
    def create_job(
        self, folder: "Folder", command: str, cores: int  # , *args: Any, **kwargs: Any
    ) -> "Job":
        raise NotImplementedError()

    @abstractmethod
    def bulk_create_jobs(self, jobs: Iterable[Dict[str, Any]]) -> List["Job"]:
        raise NotImplementedError()

    @abstractmethod
    def sync_status(self, job: "Job") -> "Job":
        raise NotImplementedError()

    @abstractmethod
    def bulk_sync_status(self, jobs: Sequence["Job"]) -> Sequence["Job"]:
        raise NotImplementedError()

    @abstractmethod
    def kill(self, job: "Job", save: bool = True) -> "Job":
        raise NotImplementedError()

    @abstractmethod
    def bulk_kill(self, jobs: Sequence["Job"]) -> Sequence["Job"]:
        raise NotImplementedError()

    @abstractmethod
    def wait_gen(
        self,
        job: Union["Job", List["Job"]],
        poll_interval: Optional[int] = None,
        timeout: Optional[int] = None,
    ) -> Iterable[List["Job"]]:
        raise NotImplementedError()

    def wait(
        self,
        job: Union["Job", List["Job"]],
        poll_interval: Optional[int] = None,
        timeout: Optional[int] = None,
        progress: bool = False,
    ) -> Optional[Iterable[List["Job"]]]:
        it = self.wait_gen(job, poll_interval=poll_interval, timeout=timeout)
        if progress:
            return it
        else:
            exhaust(it)
            return None

    @abstractmethod
    def submit(self, job: "Job", save: bool = True) -> None:
        raise NotImplementedError()

    @abstractmethod
    def bulk_submit(self, jobs: Iterable["Job"]) -> None:
        raise NotImplementedError()

    @abstractmethod
    def stdout(self, job: "Job") -> ContextManager[None]:
        raise NotImplementedError()

    @abstractmethod
    def stderr(self, job: "Job") -> ContextManager[None]:
        raise NotImplementedError()

    @abstractmethod
    def resubmit(self, job: "Job") -> "Job":
        raise NotImplementedError()

    @abstractmethod
    def bulk_resubmit(
        self, jobs: Collection["Job"], do_submit: bool = True
    ) -> Iterable["Job"]:
        raise NotImplementedError()

    @abstractmethod
    def cleanup(self, job: "Job") -> "Job":
        raise NotImplementedError()

    @abstractmethod
    def bulk_cleanup(self, jobs: Sequence["Job"]) -> Sequence["Job"]:
        raise NotImplementedError()

    @abstractmethod
    def remove(self, job: "Job") -> None:
        raise NotImplementedError()

    @abstractmethod
    def bulk_remove(self, jobs: Sequence["Job"]) -> None:
        raise NotImplementedError()

    def make_log_path(self, job: "Job") -> str:
        job_str = f"{job.job_id:>06d}"
        return os.path.abspath(
            os.path.join(self.config.jobdir, job_str[:2], job_str[2:4], job_str)
        )

    def make_output_path(self, job: "Job") -> str:
        job_str = f"{job.job_id:>06d}"
        return os.path.abspath(
            os.path.join(self.config.joboutputdir, job_str[:2], job_str[2:4], job_str)
        )



from datetime import timedelta
from unittest.mock import Mock, ANY

import pytest
import os
from schema import SchemaError

from kong.config import Config, Notifier, NotificationManager, slurm_schema


def test_config_creation(state):
    config = Config({})


def test_notifier(monkeypatch):
    notifier = Mock()
    get_notifier = Mock(return_value=notifier)
    monkeypatch.setattr("kong.config.notifiers.get_notifier", get_notifier)

    inst = Notifier("hurz")

    get_notifier.assert_called_once_with("hurz")

    inst.notify("a message", title=None)

    notifier.notify.assert_called_once_with(message=ANY)
    notifier.reset_mock()

    notifier.schema = {"properties": []}
    inst.notify("a message", title="some title")
    notifier.notify.assert_called_once_with(message=ANY)
    notifier.reset_mock()

    notifier.schema = {"properties": ["title"]}

    inst.notify("a message", title=None)
    notifier.notify.assert_called_once_with(message=ANY)
    notifier.reset_mock()

    inst.notify("a message", title="some title")
    notifier.notify.assert_called_once_with(message=ANY, title="some title")
    notifier.reset_mock()

    notifier.schema = {"properties": ["subject"]}

    inst.notify("a message", title=None)
    notifier.notify.assert_called_once_with(message=ANY)
    notifier.reset_mock()

    inst.notify("a message", title="some title")
    notifier.notify.assert_called_once_with(message=ANY, subject="some title")
    notifier.reset_mock()


def test_notificationmanager(monkeypatch):
    inst = Mock()
    inst.notify = Mock(return_value="RETURN_TAG")
    notifier = Mock(return_value=inst)
    monkeypatch.setattr("kong.config.Notifier", notifier)

    config = Mock()

    config.data = {}
    nm = NotificationManager(config)
    assert notifier.call_count == 0

    config.data = {"notify": []}
    nm = NotificationManager(config)
    assert notifier.call_count == 0

    config.data = {"notify": [{"name": "foo", "arg1": 5, "arg2": "yep"}]}

    nm = NotificationManager(config)

    notifier.assert_called_once_with(name="foo", arg1=5, arg2="yep")

    res = nm.notify("a message", "some title", "extra pos", extra_kw=42)

    inst.notify.assert_called_once_with(
        "a message", "some title", "extra pos", extra_kw=42
    )

    assert len(res) == 1
    assert res[0] == "RETURN_TAG"


def test_slurm_schema():
    assert slurm_schema.is_valid({"node_size": 2})
    assert not slurm_schema.is_valid({"node_size": -1})

    defs = slurm_schema.validate({})
    assert "sacct_delta" in defs
    assert isinstance(defs["sacct_delta"], timedelta)

    cfg = slurm_schema.validate({"sacct_delta": "10 weeks"})
    assert cfg["sacct_delta"] == timedelta(weeks=10)

    cfg = slurm_schema.validate({"sacct_delta": "50 weeks"})
    assert cfg["sacct_delta"] == timedelta(weeks=50)

    with pytest.raises(SchemaError):
        slurm_schema.validate({"sacct_delta": "blablurz"})

def test_slurm_schema_file(app_env):
    app_dir, config_path, tmp_path = app_env
    os.makedirs(app_dir)

    with open(config_path, "w") as fh:
        fh.write("""
slurm_driver:
    sacct_delta: blablurz
        """.strip())

    with pytest.raises(SchemaError):
        Config()

    with open(config_path, "w") as fh:
        fh.write("""
slurm_driver:
    sacct_delta: 10 weeks
        """.strip())

    cfg = Config()
    assert cfg.data["slurm_driver"]["sacct_delta"] == timedelta(weeks=10)

    with open(config_path, "w") as fh:
        fh.write("""
slurm_driver:
    sacct_delta: 50 weeks
        """.strip())

    cfg = Config()
    print(cfg.data)
    assert cfg.data["slurm_driver"]["sacct_delta"] == timedelta(weeks=50)



import datetime
from contextlib import contextmanager
from enum import IntFlag
from functools import wraps
from typing import Any, List, Dict, Union, cast, TYPE_CHECKING, Optional, Type, Iterator

import peewee as pw

from ..json_field import JSONField
from ..db import AutoIncrementField
from ..drivers import DriverMismatch
from ..drivers.driver_base import DriverBase
from ..config import Config
from ..model.folder import Folder
from . import BaseModel


class EnumField(pw.IntegerField):
    def __init__(self, choices: List, *args: Any, **kwargs: Any):
        self.from_db: Dict[int, str] = {int(k): k for k in choices}
        super(pw.IntegerField, self).__init__(*args, **kwargs)

    def db_value(self, value: "Job.Status") -> int:
        return int(value)

    def python_value(self, value: int) -> str:
        return self.from_db[value]


def with_driver(f: Any) -> Any:
    @wraps(f)
    def wrapper(self: "Job", *args: Any, **kwargs: Any) -> Any:
        assert (
            self._driver_instance is not None
        ), "Cannot call this method without a driver instance"
        return f(self, self._driver_instance, *args, **kwargs)

    return wrapper


class DriverField(pw.CharField):
    def __init__(self, *args: Any, **kwargs: Any):
        return super().__init__(*args, **kwargs)

    def db_value(self, value: Any) -> str:
        assert issubclass(value, DriverBase)
        class_name = ".".join([value.__module__, value.__name__])
        return class_name

    def python_value(selfself, value: str) -> Type[DriverBase]:
        import importlib

        components = value.split(".")
        module_name = ".".join(components[:-1])
        class_name = components[-1]

        module = importlib.import_module(module_name)
        class_ = getattr(module, class_name)
        return cast(Type[DriverBase], class_)


class Job(BaseModel):
    class Status(IntFlag):
        UNKNOWN = 5
        CREATED = 0
        SUBMITTED = 1
        RUNNING = 2
        FAILED = 3
        COMPLETED = 4

    class Meta:
        indexes = (
            (("batch_job_id", "driver"), True),
        )  # batch job is is unique per driver

    if TYPE_CHECKING:  # pragma: no cover
        job_id: int
        batch_job_id: str
        driver: Type[DriverBase]
        folder: "Folder"
        command: str
        data: Dict[str, Any]
        status: Status
        created_at: datetime.datetime
        updated_at: datetime.datetime
        cores: int
        memory: int
    else:
        job_id = AutoIncrementField(column_name="rowid")
        batch_job_id = pw.CharField(
            index=True, null=True
        )  # can be null, some drivers only know after submission
        # driver = EnumField(choices=drivers.__all__, null=False)
        driver = DriverField(null=False)
        folder = pw.ForeignKeyField(Folder, null=False, backref="jobs")
        command = pw.CharField(null=False)  # should allow arbitrary length in sqlite
        data = JSONField(default={})
        cores = pw.IntegerField(null=False, default=1)
        memory = pw.IntegerField(null=False, default=1000)  # memory in Megabytes
        status = EnumField(choices=Status, null=False, default=Status.CREATED)

        created_at = pw.DateTimeField(default=datetime.datetime.now)
        updated_at = pw.DateTimeField()

    _driver_instance: Optional[DriverBase] = None

    def ensure_driver_instance(self, arg: Union[DriverBase, Config]) -> None:
        if self._driver_instance is not None:
            return
        if isinstance(arg, Config):
            self._driver_instance = self.driver(arg)
        else:
            if not isinstance(arg, self.driver):
                raise DriverMismatch(
                    f"Given driver {arg} is not instance of {self.driver}"
                )
            self._driver_instance = arg

    @property
    def driver_instance(self) -> DriverBase:
        assert self._driver_instance is not None
        return self._driver_instance

    def save(self, *args: Any, **kwargs: Any) -> None:
        # assert self.driver in drivers.__all__, f"{self.driver} is not a valid driver"
        assert self.command is not None, "Need to specify a command"
        assert len(self.command) > 0, "Command must be longer than 0"
        self.updated_at = datetime.datetime.now()
        super().save(*args, **kwargs)

    @with_driver
    def remove(self, driver: DriverBase) -> None:
        driver.remove(self)

    @property
    def log_dir(self) -> str:
        return str(self.data["log_dir"])

    @property
    def output_dir(self) -> str:
        return str(self.data["output_dir"])

    @with_driver
    def submit(self, driver: DriverBase) -> Any:
        driver.submit(self)

    @with_driver
    def resubmit(self, driver: DriverBase) -> Any:
        driver.resubmit(self)

    @with_driver
    def wait(self, driver: DriverBase, timeout: Optional[int] = None) -> None:
        driver.wait(self, timeout=timeout)

    @with_driver
    def kill(self, driver: DriverBase) -> None:
        driver.kill(self)

    @with_driver
    def get_status(self, driver: DriverBase) -> Status:  # noqa: F821
        self.reload()
        driver.sync_status(self)
        return self.status

    @with_driver  # type: ignore
    @contextmanager  # type: ignore
    def stdout(self, driver: DriverBase) -> Iterator[None]:
        with driver.stdout(self) as fh:
            yield fh

    @with_driver  # type: ignore
    @contextmanager  # type: ignore
    def stderr(self, driver: DriverBase) -> Iterator[None]:
        with driver.stderr(self) as fh:
            yield fh

    def __str__(self) -> str:
        return f"Job<{self.job_id}, {self.batch_job_id}, {str(self.status)}>"


color_dict = {
    Job.Status.UNKNOWN: "red",
    Job.Status.CREATED: "white",
    Job.Status.SUBMITTED: "yellow",
    Job.Status.RUNNING: "blue",
    Job.Status.FAILED: "red",
    Job.Status.COMPLETED: "green",
}



import os
import random
import shutil
import time

import psutil
import pytest
from unittest.mock import Mock, call

from kong.drivers import DriverMismatch, InvalidJobStatus
from kong.drivers.local_driver import LocalDriver
import kong
from kong.model import Folder, Job
from conftest import skip_lxplus


@pytest.fixture
def state(app_env, db, monkeypatch):
    app_dir, config_path, tmp_path = app_env
    with monkeypatch.context() as m:
        m.setattr(
            "click.prompt",
            Mock(
                side_effect=[
                    "kong.drivers.local_driver.LocalDriver",
                    os.path.join(app_dir, "joblog"),
                    os.path.join(app_dir, "joboutput"),
                ]
            ),
        )
        kong.setup.setup(None)
    return kong.get_instance()


@pytest.fixture
def driver(state):
    return LocalDriver(state.config)


def test_create_job(driver, tree, state):
    with pytest.raises(AssertionError):
        driver.create_job(tree, command="sleep 1", batch_job_id=42)
    with pytest.raises(AssertionError):
        driver.create_job(tree, command="sleep 1", driver="OtherDriver")

    config = state.config
    assert len(os.listdir(config.jobdir)) == 0, "Job dir is not empty"

    j1 = driver.create_job(tree, command="sleep 1")
    assert j1 is not None
    assert j1.folder == tree
    assert len(tree.jobs) == 1
    assert tree.jobs[0] == j1
    assert j1.status == Job.Status.CREATED

    assert os.path.exists(j1.data["log_dir"]), "Does not create job directory"

    assert os.path.exists(j1.data["output_dir"]), "Does not create job output directory"

    assert os.path.isfile(
        os.path.join(j1.data["log_dir"], "jobscript.sh")
    ), "Does not create job script"

    f2 = tree.subfolder(("f2"))
    j2 = driver.create_job(f2, command="sleep 1")
    assert j2 is not None
    assert j2.folder == f2
    assert len(f2.jobs) == 1
    assert f2.jobs[0] == j2
    assert j2.status == Job.Status.CREATED
    assert os.path.exists(j2.data["log_dir"]), "Does not create job directory"
    assert os.path.exists(j2.data["output_dir"]), "Does not create job output directory"
    assert os.path.isfile(
        os.path.join(j2.data["log_dir"], "jobscript.sh")
    ), "Does not create job script"

    print(j2.data)

    assert j1.batch_job_id != j2.batch_job_id


class ValidDriver(kong.drivers.driver_base.DriverBase):
    def __init__(self):
        pass


def test_driver_mismatch(driver, state, monkeypatch):
    root = Folder.get_root()

    monkeypatch.setattr(
        "kong.drivers.driver_base.DriverBase.__abstractmethods__", set()
    )

    j1 = Job.create(folder=root, command="sleep 1", driver=ValidDriver)
    assert j1.driver == ValidDriver

    # updating with local driver does not work
    with pytest.raises(DriverMismatch):
        driver.sync_status(j1)

    j2 = Job.get(j1.job_id)
    with pytest.raises(DriverMismatch):
        j2.ensure_driver_instance(driver)


def test_job_rm_cleans_up(driver, state):
    j1 = driver.create_job(command="sleep 1", folder=state.cwd)
    assert j1 is not None
    assert os.path.exists(j1.data["log_dir"]), "Does not create job directory"
    assert os.path.exists(j1.data["output_dir"]), "Does not create output directory"
    assert os.path.exists(j1.data["scratch_dir"]), "Does not create scratch directory"

    driver.remove(j1)

    assert not os.path.exists(
        j1.data["log_dir"]
    ), "Driver does not cleanup job directory"
    assert not os.path.exists(
        j1.data["output_dir"]
    ), "Driver does not cleanup output directory"
    assert not os.path.exists(
        j1.data["scratch_dir"]
    ), "Does not cleanup scratch directory"

    j2 = driver.create_job(command="sleep 1", folder=state.cwd)
    assert j2 is not None
    assert os.path.exists(j2.data["log_dir"]), "Does not create job directory"
    assert os.path.exists(j2.data["output_dir"]), "Does not create output directory"
    assert os.path.exists(j2.data["scratch_dir"]), "Does not create scratch directory"

    shutil.rmtree(j2.data["log_dir"])
    assert not os.path.exists(j2.data["log_dir"]), "Does not create job directory"

    driver.remove(j2)

    assert not os.path.exists(
        j2.data["log_dir"]
    ), "Driver does not cleanup job directory"
    assert not os.path.exists(
        j2.data["output_dir"]
    ), "Driver does not cleanup output directory"
    assert not os.path.exists(
        j2.data["scratch_dir"]
    ), "Does not cleanup scratch directory"


def test_job_bulk_remove(driver, state):
    jobs = [
        driver.create_job(command="sleep 1", folder=state.cwd),
        driver.create_job(command="sleep 1", folder=state.cwd),
        driver.create_job(command="sleep 1", folder=state.cwd),
    ]
    for job in jobs:
        assert os.path.exists(job.data["log_dir"]), "Does not create job directory"
        assert os.path.exists(
            job.data["output_dir"]
        ), "Does not create output directory"
        assert os.path.exists(
            job.data["scratch_dir"]
        ), "Does not create scratch directory"

    driver.bulk_remove(jobs)

    for job in jobs:
        assert not os.path.exists(
            job.data["log_dir"]
        ), "Driver does not cleanup job directory"
        assert not os.path.exists(
            job.data["output_dir"]
        ), "Driver does not cleanup output directory"
        assert not os.path.exists(
            job.data["scratch_dir"]
        ), "Does not cleanup scratch directory"


def test_job_cleanup_status(driver, state):
    j1 = driver.create_job(command="sleep 1", folder=state.cwd)
    assert j1 is not None
    assert os.path.exists(j1.data["log_dir"])
    assert os.path.exists(j1.data["output_dir"])
    j1.status = Job.Status.RUNNING
    j1.save()
    with pytest.raises(InvalidJobStatus):
        driver.cleanup(j1)
    assert os.path.exists(j1.data["log_dir"])
    assert os.path.exists(j1.data["output_dir"])
    j1.status = Job.Status.SUBMITTED
    j1.save()
    with pytest.raises(InvalidJobStatus):
        driver.cleanup(j1)
    assert os.path.exists(j1.data["log_dir"])
    assert os.path.exists(j1.data["output_dir"])

    for status in [
        Job.Status.CREATED,
        Job.Status.FAILED,
        Job.Status.COMPLETED,
        Job.Status.UNKNOWN,
    ]:
        j = driver.create_job(command="sleep 1", folder=state.cwd)
        assert j is not None
        j.status = status
        j.save()
        assert os.path.exists(j.data["log_dir"])
        assert os.path.exists(j.data["output_dir"])
        driver.cleanup(j)
        assert not os.path.exists(j.data["log_dir"])
        assert not os.path.exists(j.data["output_dir"])


def test_job_bulk_cleanup(driver, state):
    jobs = [
        driver.create_job(command="sleep 1", folder=state.cwd),
        driver.create_job(command="sleep 1", folder=state.cwd),
        driver.create_job(command="sleep 1", folder=state.cwd),
    ]

    for job in jobs:
        assert os.path.exists(job.data["log_dir"])
        assert os.path.exists(job.data["output_dir"])

    jobs[0].status = Job.Status.RUNNING
    jobs[0].save()
    with pytest.raises(InvalidJobStatus):
        driver.bulk_cleanup(jobs)

    for job in jobs:
        assert os.path.exists(job.data["log_dir"])
        assert os.path.exists(job.data["output_dir"])

    jobs[0].status = Job.Status.CREATED
    jobs[0].save()

    driver.bulk_cleanup(jobs)

    for job in jobs:
        assert not os.path.exists(job.data["log_dir"])
        assert not os.path.exists(job.data["output_dir"])


def test_job_env_is_valid(driver, state):
    root = Folder.get_root()

    def run_get_env(**kwargs):
        j1 = driver.create_job(folder=root, command="env", **kwargs)
        j1.submit()
        j1.wait()
        env = {}
        with j1.stdout() as fh:
            raw = fh.read().strip().split("\n")
            for line in raw:
                if "=" not in line:
                    continue
                k, v = line.split("=", 1)
                env[k] = v
        return j1, env

    job, env = run_get_env()
    output_dir = driver.make_output_path(job)
    log_dir = driver.make_log_path(job)

    assert env["KONG_JOB_NPROC"] == "1"
    assert env["KONG_JOB_SCRATCHDIR"] != ""
    assert os.path.exists(env["KONG_JOB_SCRATCHDIR"])
    assert env["KONG_JOB_ID"] == str(job.job_id)
    assert (
        env["KONG_JOB_OUTPUT_DIR"] == output_dir
        and job.data["output_dir"] == output_dir
    )
    assert env["KONG_JOB_LOG_DIR"] == log_dir and job.data["log_dir"] == log_dir

    job, env = run_get_env(cores=8)
    output_dir = driver.make_output_path(job)
    log_dir = driver.make_log_path(job)

    assert env["KONG_JOB_NPROC"] == "8"
    assert env["KONG_JOB_SCRATCHDIR"] != ""
    assert os.path.exists(env["KONG_JOB_SCRATCHDIR"])
    assert env["KONG_JOB_ID"] == str(job.job_id)
    assert (
        env["KONG_JOB_OUTPUT_DIR"] == output_dir
        and job.data["output_dir"] == output_dir
    )
    assert env["KONG_JOB_LOG_DIR"] == log_dir and job.data["log_dir"] == log_dir


def test_run_job(driver, state, db):
    root = Folder.get_root()

    value = "I AM THE EXPECTED OUTPUT"
    script = f"echo '{value}'"

    j1 = driver.create_job(command=script, folder=root)
    assert j1.status == Job.Status.CREATED

    driver.submit(j1)
    assert j1.status == Job.Status.SUBMITTED

    print("WAIT")
    driver.wait(j1, timeout=2)
    print("WAIT DONE")
    driver.wait(j1, timeout=2)
    assert j1.status == Job.Status.COMPLETED

    with driver.stdout(j1) as so:
        out = so.read().strip()

    assert out == value


def test_submit_invalid_status(driver, state):
    j1 = driver.create_job(command="sleep 1", folder=state.cwd)
    j1.data["pid"] = 123
    for status in (
        Job.Status.COMPLETED,
        Job.Status.FAILED,
        Job.Status.RUNNING,
        Job.Status.UNKNOWN,
    ):
        j1.status = status
        j1.save()
        with pytest.raises(InvalidJobStatus):
            driver.submit(j1)


def test_run_stdout_stderr(driver, state):
    root = Folder.get_root()
    error = "ERRORERROR"
    value = "VALUEVALUE"

    j1 = driver.create_job(command=f"echo '{error}' 1>&2 ; echo '{value}'", folder=root)
    j1.submit()
    j1.wait()
    assert j1.status == Job.Status.COMPLETED

    # os.system("ls -al "+os.path.join(state.config.jobdir, os.listdir(state.config.jobdir)[0]))
    # os.system("cat "+os.path.join(state.config.jobdir, os.listdir(state.config.jobdir)[0], "stdout.txt"))
    # os.system("cat "+os.path.join(state.config.jobdir, os.listdir(state.config.jobdir)[0], "stderr.txt"))

    with j1.stderr() as fh:
        assert fh.read().strip() == error
    with j1.stdout() as fh:
        assert fh.read().strip() == value


def test_stdout_stderr_invalid_status(driver, state, monkeypatch):
    monkeypatch.setattr(driver, "sync_status", Mock())
    j1 = driver.create_job(command="sleep 1", folder=state.cwd)
    for status in (
        Job.Status.CREATED,
        Job.Status.SUBMITTED,
        Job.Status.RUNNING,
        Job.Status.UNKNOWN,
    ):
        j1.status = status
        j1.save()
        with pytest.raises(InvalidJobStatus):
            with driver.stdout(j1):
                pass

        with pytest.raises(InvalidJobStatus):
            with driver.stderr(j1):
                pass


def test_run_job_already_completed(driver, state):
    root = Folder.get_root()
    j1 = driver.create_job(command="echo 'hi'", folder=root)
    j1.submit()

    # already waited, process is reaped
    try:
        print(j1.data["pid"])
        proc = psutil.Process(j1.data["pid"])
        print(proc.status())
        proc.wait()
    except psutil.NoSuchProcess:
        # weird, but ok
        pass

    driver.wait(j1)
    assert j1.status == Job.Status.COMPLETED


@skip_lxplus
def test_run_job_timeout(driver, state):
    root = Folder.get_root()
    j1 = driver.create_job(command="sleep 1", folder=root)
    j1.submit()

    with pytest.raises(TimeoutError):
        driver.wait(j1, timeout=0.1)
    assert j1.status == Job.Status.RUNNING
    time.sleep(1)
    driver.wait(j1, timeout=0.1)
    assert j1.status == Job.Status.COMPLETED


def test_run_failed(driver, state):
    root = Folder.get_root()
    j1 = driver.create_job(command="exit 1", folder=root)
    j2 = driver.create_job(command="exit 127", folder=root)

    j1.submit()
    j2.submit()

    j1.wait()
    j2.wait()

    assert j1.status == Job.Status.FAILED
    assert j1.data["exit_code"] == 1

    assert j2.status == Job.Status.FAILED
    assert j2.data["exit_code"] == 127


@skip_lxplus
def test_run_killed(driver, state):
    root = Folder.get_root()
    j1 = driver.create_job(command="sleep 10", folder=root)
    j1.submit()
    proc = psutil.Process(pid=j1.data["pid"])
    proc.kill()
    j1.wait()
    assert j1.status == Job.Status.UNKNOWN


@skip_lxplus
def test_run_terminated(driver, state):
    root = Folder.get_root()
    j1 = driver.create_job(command="echo 'begin'; sleep 10 ; echo 'end'", folder=root)
    j1.submit()
    proc = psutil.Process(pid=j1.data["pid"])
    time.sleep(0.2)
    driver.sync_status(j1)
    assert j1.status == Job.Status.RUNNING
    for child in proc.children(
        recursive=True
    ):  # or parent.children() for recursive=False
        child.terminate()
    proc.terminate()
    j1.wait()
    assert j1.status == Job.Status.FAILED


def test_run_kill(driver, state):
    root = Folder.get_root()
    j1 = driver.create_job(command="echo 'begin'; sleep 10 ; echo 'end'", folder=root)

    driver.kill(j1)
    assert j1.status == Job.Status.FAILED

    j1.status = Job.Status.CREATED
    j1.submit()
    driver.kill(j1)
    j1.wait()
    assert j1.status == Job.Status.FAILED

    j2 = driver.create_job(command="echo 'begin'; sleep 10 ; echo 'end'", folder=root)
    j2.submit()
    time.sleep(0.2)  # wait a bit until running
    assert j2.get_status() == Job.Status.RUNNING
    j2.kill()
    assert j2.status == Job.Status.FAILED  # should be failed right away
    j2.wait()
    assert j2.status == Job.Status.FAILED  # shouldn't change after waiting


@skip_lxplus
def test_bulk_submit(driver, state):
    root = Folder.get_root()

    jobs = [
        driver.create_job(folder=root, command=f"sleep 0.1; echo 'JOB{i}'")
        for i in range(15)
    ]

    for job in jobs:
        assert job.status == Job.Status.CREATED

    driver.bulk_submit(jobs)

    for job in jobs:
        assert job.status == Job.Status.SUBMITTED

    driver.wait(jobs)

    for job in jobs:
        assert job.status == Job.Status.COMPLETED


def test_bulk_create(driver, state):
    root = Folder.get_root()

    jobs = driver.bulk_create_jobs(
        [dict(folder=root, command=f"sleep 0.1; echo 'JOB{i}'") for i in range(15)]
    )

    for job in jobs:
        assert job.status == Job.Status.CREATED


def test_bulk_kill(driver, state):
    root = Folder.get_root()

    jobs = driver.bulk_create_jobs(
        [dict(folder=root, command=f"sleep 10; echo 'JOB{i}'") for i in range(15)]
    )

    for job in jobs:
        assert job.status == Job.Status.CREATED

    driver.bulk_submit(jobs)

    for job in jobs:
        assert job.status == Job.Status.SUBMITTED

    driver.bulk_kill(jobs)

    for job in jobs:
        assert job.status == Job.Status.FAILED


@skip_lxplus
def test_bulk_wait(driver, state):
    root = Folder.get_root()

    jobs = []
    for i in range(15):
        job = driver.create_job(
            folder=root, command=f"sleep {random.random()} ; echo 'JOB{i}'"
        )
        job.submit()
        jobs.append(job)

    sjobs = len(jobs)

    for i in range(15):
        job = driver.create_job(
            folder=root,
            command=f"sleep {random.random()} ; echo 'JOB{i+sjobs}' 1>&2 ; exit 1",
        )
        job.submit()
        jobs.append(job)

    driver.wait(jobs)

    for i, job in enumerate(jobs[:15]):
        assert job.status == Job.Status.COMPLETED
        with job.stdout() as fh:
            assert fh.read().strip() == f"JOB{i}"
    for i, job in enumerate(jobs[15:]):
        assert job.status == Job.Status.FAILED
        with job.stderr() as fh:
            assert fh.read().strip() == f"JOB{i+sjobs}"


def test_sync_status(driver, state, monkeypatch, tmpdir):
    root = Folder.get_root()
    j1 = driver.create_job(
        command="echo 'begin'; sleep 0.2 ; echo 'end' ; exit 1", folder=root
    )
    j1.data["pid"] = 123
    j1.status = Job.Status.RUNNING
    j1.save()

    proc = Mock()
    proc.is_running = Mock(return_value=False)
    monkeypatch.setattr("psutil.Process", Mock(return_value=proc))

    exit_status_file = tmpdir.join("exitcode.txt")
    exit_status_file.write("0")

    j1.data["exit_status_file"] = str(exit_status_file)
    j1u = driver.sync_status(j1)
    assert j1u.status == Job.Status.COMPLETED

    j1.status = Job.Status.RUNNING
    j1.save()

    exit_status_file.write("1")
    j1u = driver.sync_status(j1)
    assert j1u.status == Job.Status.FAILED


@skip_lxplus
def test_bulk_sync(driver, state):
    root = Folder.get_root()

    jobs = []
    for i in range(15):
        job = driver.create_job(
            folder=root, command=f"sleep {0.1 + random.random()*0.2} ; echo 'JOB{i}'"
        )
        job.submit()
        jobs.append(job)

    sjobs = len(jobs)

    for i in range(15):
        job = driver.create_job(
            folder=root,
            command=f"sleep {0.1 + random.random()*0.2} ; echo 'JOB{i+sjobs}' 1>&2 ; exit 1",
        )
        job.submit()
        jobs.append(job)

    time.sleep(0.4)
    driver.bulk_sync_status(jobs)

    for i, job in enumerate(jobs[:15]):
        assert job.status == Job.Status.COMPLETED
        with job.stdout() as fh:
            assert fh.read().strip() == f"JOB{i}"
    for i, job in enumerate(jobs[15:]):
        assert job.status == Job.Status.FAILED
        with job.stderr() as fh:
            assert fh.read().strip() == f"JOB{i+sjobs}"


@skip_lxplus
def test_job_resubmit(driver, state, monkeypatch):
    root = Folder.get_root()
    j1 = driver.create_job(
        command="echo 'begin'; sleep 0.2 ; echo 'end' ; exit 1", folder=root
    )
    j1.submit()
    j1.wait()
    assert j1.status == Job.Status.FAILED
    with j1.stdout() as fh:
        assert fh.read().strip() == "begin\nend"

    # we need to prevent driver from actually calling submit
    submit = Mock()
    with monkeypatch.context() as m:
        m.setattr(driver, "submit", submit)
        makedirs = Mock()
        m.setattr("os.makedirs", makedirs)
        remove = Mock(wraps=os.remove)
        m.setattr("os.remove", remove)
        driver.resubmit(j1)
        assert makedirs.call_count == 2
    submit.assert_called_once()
    remove.assert_has_calls(
        [call(j1.data[p]) for p in ["exit_status_file", "stdout", "stderr"]]
    )

    # now actually hit submit
    driver.submit(j1)

    assert j1.status == Job.Status.SUBMITTED
    j1.wait()
    assert j1.status == Job.Status.FAILED


def test_job_resubmit_already_deleted(driver, state, monkeypatch):
    root = Folder.get_root()
    j1 = driver.create_job(
        command="echo 'begin'; sleep 0.2 ; echo 'end' ; exit 1", folder=root
    )
    j1.status = Job.Status.COMPLETED
    j1.save()

    # it never ran, so these shouldn't exist
    for path in ["exit_status_file", "stdout", "stderr"]:
        assert not os.path.exists(j1.data[path])
    for d in ["scratch_dir", "output_dir"]:
        path = j1.data[d]
        shutil.rmtree(path)

    submit = Mock()
    with monkeypatch.context() as m:
        m.setattr(driver, "submit", submit)
        makedirs = Mock()
        m.setattr("os.makedirs", makedirs)
        remove = Mock()
        m.setattr("os.remove", remove)
        driver.resubmit(j1)
        assert makedirs.call_count == 0
        assert remove.call_count == 0
    submit.assert_called_once()
    for path in ["exit_status_file", "stdout", "stderr"]:
        assert not os.path.exists(j1.data[path])


def test_resubmit_invalid_status(driver, state, monkeypatch):
    monkeypatch.setattr(driver, "sync_status", Mock())
    j1 = driver.create_job(command="sleep 1", folder=state.cwd)
    for status in (Job.Status.CREATED, Job.Status.SUBMITTED, Job.Status.RUNNING):
        j1.status = status
        j1.save()
        with pytest.raises(InvalidJobStatus):
            driver.resubmit(j1)


@skip_lxplus
def test_job_bulk_resubmit(driver, state, monkeypatch):
    root = Folder.get_root()

    jobs = [
        driver.create_job(
            command="echo 'begin'; sleep 0.2 ; echo 'end' ; exit 1", folder=root
        ),
        driver.create_job(
            command="echo 'begin'; sleep 0.2 ; echo 'end' ; exit 1", folder=root
        ),
        driver.create_job(
            command="echo 'begin'; sleep 0.2 ; echo 'end' ; exit 1", folder=root
        ),
    ]

    for d in ["scratch_dir", "output_dir"]:
        path = jobs[0].data[d]
        shutil.rmtree(path)

    jobs[0].status = Job.Status.FAILED
    jobs[0].save()

    driver.bulk_submit(jobs[1:])
    driver.wait(jobs)

    for job in jobs[1:]:
        assert job.status == Job.Status.FAILED
        with job.stdout() as fh:
            assert fh.read().strip() == "begin\nend"

    # we need to prevent driver from actually calling submit
    submit = Mock()
    makedirs = Mock()
    remove = Mock(wraps=os.remove)
    with monkeypatch.context() as m:
        m.setattr(driver, "submit", submit)
        m.setattr("os.makedirs", makedirs)
        m.setattr("os.remove", remove)
        driver.bulk_resubmit(jobs)
    assert submit.call_count == len(jobs)
    remove.assert_has_calls(
        sum(
            [
                [call(j.data[p]) for p in ["exit_status_file", "stdout", "stderr"]]
                for j in jobs[1:]
            ],
            [],
        )
    )
    makedirs.assert_has_calls(
        sum(
            [
                [call(j.data[p]) for p in ["scratch_dir", "output_dir"]]
                for j in jobs[1:]
            ],
            [],
        )
    )

    for job in jobs:
        for path in ["exit_status_file", "stdout", "stderr"]:
            assert not os.path.exists(job.data[path])

    # now actually hit submit
    driver.bulk_submit(jobs)

    for job in jobs:
        assert job.status == Job.Status.SUBMITTED
    driver.wait(jobs)
    for job in jobs:
        assert job.status == Job.Status.FAILED


def test_resubmit_bulk_invalid_status(driver, state, monkeypatch):
    monkeypatch.setattr(driver, "sync_status", Mock())
    j1 = driver.create_job(command="sleep 1", folder=state.cwd)
    for status in (Job.Status.CREATED, Job.Status.SUBMITTED, Job.Status.RUNNING):
        j1.status = status
        j1.save()
        with pytest.raises(InvalidJobStatus):
            driver.bulk_resubmit([j1])



import datetime
import os
import re
from abc import ABC, abstractmethod
from datetime import date, timedelta
from typing import Iterator, Iterable, Optional, Union, List, Sequence, cast, Collection

import sh
from jinja2 import Environment, DictLoader

from kong.drivers.batch_driver_base import BatchDriverBase
from ..drivers import InvalidJobStatus
from ..logger import logger
from ..config import Config
from ..db import database
from ..model import Job, Folder
from ..util import make_executable, format_timedelta, parse_timedelta
from .driver_base import checked_job


class SlurmAccountingItem:
    job_id: int
    status: Job.Status
    exit_code: int

    def __init__(self, job_id: int, status: Job.Status, exit_code: int):
        self.job_id = job_id
        self.status = status
        self.exit_code = exit_code

    @classmethod
    def from_parts(
        cls, job_id: str, status_str: str, exit: str
    ) -> "SlurmAccountingItem":
        exit_code, _ = exit.split(":", 1)

        if status_str == "PENDING":
            status = Job.Status.SUBMITTED
        elif status_str == "COMPLETED":
            status = Job.Status.COMPLETED
        elif status_str == "FAILED" or status_str.startswith("CANCELLED"):
            status = Job.Status.FAILED
        elif status_str == "RUNNING":
            status = Job.Status.RUNNING
        else:
            status = Job.Status.UNKNOWN

        return cls(int(job_id), status, int(exit_code))

    def __repr__(self) -> str:
        return f"SAI<{self.job_id}, {self.status}, {self.exit_code}>"

    def __eq__(self, other: object) -> bool:
        assert isinstance(other, self.__class__)
        return (
            self.job_id == other.job_id
            and self.status == other.status
            and self.exit_code == other.exit_code
        )


class SlurmInterface(ABC):
    @abstractmethod
    def sacct(
        self, jobs: Collection["Job"], start_delta: timedelta
    ) -> Iterator[SlurmAccountingItem]:
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def sbatch(self, job: "Job") -> int:
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def scancel(self, job: "Job") -> None:
        raise NotImplementedError()  # pragma: no cover


class ShellSlurmInterface(SlurmInterface):
    _sacct: Optional[sh.Command] = None
    _sbatch: Optional[sh.Command] = None
    _scancel: Optional[sh.Command] = None

    subreg = re.compile(r".* (\d*)$")

    def __init__(self) -> None:  # pragma: no cover
        self._sacct = sh.Command("sacct")
        self._sbatch = sh.Command("sbatch")
        self._scancel = sh.Command("scancel")

    def sacct(
        self, jobs: Collection["Job"], start_delta: timedelta
    ) -> Iterator[SlurmAccountingItem]:

        logger.debug("Getting job info for %d jobs", len(jobs))
        starttime = date.today() - start_delta

        args = dict(
            brief=True, noheader=True, parsable2=True, starttime=starttime, _iter=True
        )

        if len(jobs) > 0 and len(jobs) < 20:
            if all(j.batch_job_id is None for j in jobs):
                logger.debug("no jobs given that are known to the scheduder")
                return []

            job_ids = ",".join(
                [str(j.batch_job_id) for j in jobs if j.batch_job_id is not None]
            )
            args["jobs"] = job_ids

            logger.debug("Job argument: %s", job_ids)

        assert self._sacct is not None
        for line in self._sacct(**args):
            job_id, status, exit = line.split("|", 3)
            if not job_id.isdigit():
                continue
            yield SlurmAccountingItem.from_parts(job_id, status, exit)

    def sbatch(self, job: Job) -> int:
        assert self._sbatch is not None
        raw = self._sbatch(job.data["batchfile"])
        logger.debug("sbatch: %s", raw)
        m = self.subreg.match(str(raw))
        assert m is not None
        return int(m.group(1))

    def scancel(self, job: Job) -> None:
        assert self._scancel is not None
        res = self._scancel(job.batch_job_id)
        logger.debug("scancel: %s", res)


jobscript_tpl_str = """
#!/usr/bin/env bash

export KONG_JOB_ID={{internal_job_id}}
export KONG_JOB_OUTPUT_DIR={{output_dir}}
export KONG_JOB_LOG_DIR={{log_dir}}
export KONG_JOB_NPROC={{cores}}
export KONG_JOB_SCRATCHDIR=/localscratch/${{ "{" }}SLURM_JOB_ID{{ "}" }}/

mkdir -p $KONG_JOB_SCRATCHDIR

stdout={{stdout}}

({{command}}) > $stdout 2>&1
""".strip()

batchfile_tpl_str = """
#!/bin/bash
#SBATCH -J {{name}}
#SBATCH -o {{slurm_out}}
#SBATCH -p {{queue}}
              
#SBATCH -n {{ntasks}}
#SBATCH -N {{nnodes}}
#SBATCH -c {{cores}}
#SBATCH --mem-per-cpu {{memory}}M 
#SBATCH -t {{walltime}}
{%- if licenses is not none %}
#SBATCH -L {{licenses}}
{% endif %}

#SBATCH -A {{account}}

srun --export=NONE {{jobscript}}
""".strip()  # noqa: W291, W293

env = Environment(
    loader=DictLoader({"batchfile": batchfile_tpl_str, "jobscript": jobscript_tpl_str})
)

batchfile_tpl = env.get_template("batchfile")
jobscript_tpl = env.get_template("jobscript")


class SlurmDriver(BatchDriverBase):
    slurm: SlurmInterface

    def __init__(self, config: Config, slurm: Optional[SlurmInterface] = None):
        self.slurm = slurm or ShellSlurmInterface()
        super().__init__(config)
        self.slurm_config = self.config.data["slurm_driver"]

    def create_job(
        self,
        folder: "Folder",
        command: str,
        cores: int = 1,
        memory: int = 1000,
        nnodes: int = 1,
        ntasks: int = 1,
        queue: Optional[str] = None,
        name: Optional[str] = None,
        walltime: Union[timedelta, str] = timedelta(minutes=30),
        licenses: Optional[str] = None,
    ) -> "Job":

        if queue is None:
            queue = self.slurm_config["default_queue"]

        job: Job = Job.create(
            folder=folder,
            batch_job_id=None,  # don't have one until submission
            command=command,
            driver=self.__class__,
            cores=cores,
            memory=memory,
        )

        if name is None:
            name = f"kong_job_{job.job_id}"

        # in job dir, create output dir
        output_dir = self.make_output_path(job)
        os.makedirs(output_dir, exist_ok=True)

        log_dir = self.make_log_path(job)
        os.makedirs(log_dir, exist_ok=True)

        stdout = os.path.abspath(os.path.join(log_dir, "stdout.txt"))
        slurm_out = os.path.abspath(os.path.join(log_dir, "slurm_out.txt"))

        batchfile = os.path.join(log_dir, "batchfile.sh")
        jobscript = os.path.join(log_dir, "jobscript.sh")

        if isinstance(walltime, str):
            norm_walltime = format_timedelta(parse_timedelta(walltime))
        elif isinstance(walltime, timedelta):
            norm_walltime = format_timedelta(walltime)
        else:
            raise ValueError("Walltime must be timedelta or string")

        job.data = dict(
            stdout=stdout,
            slurm_out=slurm_out,
            jobscript=jobscript,
            batchfile=batchfile,
            output_dir=output_dir,
            log_dir=log_dir,
            name=name,
            queue=queue,
            nnodes=nnodes,
            ntasks=ntasks,
            exit_code=0,
            walltime=norm_walltime,
            account=self.config.slurm_driver["account"],
            licenses=licenses,
        )
        job.save()

        values = dict(
            batchfile=batchfile,
            jobscript=jobscript,
            command=command,
            stdout=stdout,
            slurm_out=slurm_out,
            internal_job_id=job.job_id,
            log_dir=log_dir,
            output_dir=output_dir,
            cores=cores,
            nnodes=nnodes,
            ntasks=ntasks,
            memory=memory,
            account=self.config.slurm_driver["account"],
            name=name,
            queue=queue,
            walltime=norm_walltime,
            licenses=licenses,
        )

        batchfile_content = batchfile_tpl.render(**values)

        with open(batchfile, "w") as fh:
            fh.write(batchfile_content)

        jobscript_content = jobscript_tpl.render(**values)
        with open(job.data["jobscript"], "w") as fh:
            fh.write(jobscript_content)

        make_executable(job.data["jobscript"])

        job._driver_instance = self
        return job

    def bulk_sync_status(self, jobs: Collection["Job"]) -> Sequence["Job"]:
        logger.debug("Bulk sync status with %d jobs", len(jobs))
        for job in jobs:
            self._check_driver(job)

        now = datetime.datetime.now()

        def proc() -> Iterable[Job]:
            job_not_found = 0
            for item in self.slurm.sacct(jobs, self.slurm_config["sacct_delta"]):
                job = Job.get_or_none(batch_job_id=item.job_id)
                if job is None:
                    job_not_found += 1
                    continue
                job.status = item.status
                job.data["exit_code"] = item.exit_code
                job.updated_at = now
                assert job.status != Job.Status.CREATED, "Job updated to created?"
                yield job
            if job_not_found > 0:
                logger.info(
                    "Tried to fetch %d jobs which where not found in the database",
                    job_not_found,
                )

        with database.atomic():
            Job.bulk_update(
                proc(),
                fields=[Job.data, Job.status, Job.updated_at],
                batch_size=self.batch_size,
            )
        # reload updated jobs
        ids = [j.job_id for j in jobs]
        logger.debug(
            "Going to reload %d jobs, batch size %d", len(ids), self.select_batch_size
        )

        fetched: List[Job] = list(
            Job.bulk_select(Job.job_id, ids, batch_size=self.select_batch_size)
        )
        return cast(Sequence[Job], fetched)

    @checked_job
    def kill(self, job: "Job", save: bool = True) -> None:
        if job.status in (Job.Status.CREATED, Job.Status.UNKNOWN):
            logger.debug("Job %s in %s, simply setting to failed", job, job.status)
            job.status = Job.Status.FAILED
        elif job.status in (Job.Status.RUNNING, Job.Status.SUBMITTED):
            self.slurm.scancel(job)
            job.status = Job.Status.FAILED
        else:
            logger.debug("Job %s in %s, do nothing")
        if save:
            job.save()

    @checked_job
    def submit(self, job: "Job", save: bool = True) -> None:
        if job.status != Job.Status.CREATED:
            raise InvalidJobStatus(f"Cannot submit job {job} in status {job.status}")
        job.batch_job_id = str(self.slurm.sbatch(job))
        job.status = Job.Status.SUBMITTED

        if save:
            job.save()



import os
import time
import uuid
from conftest import skip_lxplus

import pytest
from unittest.mock import Mock, MagicMock, ANY
import peewee as pw

import kong
from kong.config import Config
from kong.drivers.local_driver import LocalDriver
from kong.model import Folder, Job
import kong.drivers
from kong.state import DoesNotExist, CannotCreateError, CannotRemoveIsFolder
from kong.util import exhaust


@pytest.fixture
def cfg(app_env, monkeypatch):
    app_dir, config_path, tmp_path = app_env
    with monkeypatch.context() as m:
        m.setattr(
            "click.prompt",
            Mock(
                side_effect=[
                    "kong.drivers.local_driver.LocalDriver",
                    os.path.join(app_dir, "joblog"),
                    os.path.join(app_dir, "joboutput"),
                ]
            ),
        )
        kong.setup.setup(None)
    _cfg = kong.config.Config()
    return _cfg


def test_init(cfg, db):
    # this requires a database to be created externally
    s = kong.state.State(cfg, kong.model.Folder.get_root())
    assert s is not None


def test_get_instance(cfg, monkeypatch):
    orig_init = kong.db.database.init
    init = Mock(side_effect=lambda _: orig_init(":memory:"))
    monkeypatch.setattr("kong.db.database.init", init)
    s = kong.state.State.get_instance()
    assert s is not None
    init.assert_called_once()
    assert s.cwd == kong.model.Folder.get_root()


def test_module_get_instance(cfg, monkeypatch):
    orig_init = kong.db.database.init
    init = Mock(side_effect=lambda _: orig_init(":memory:"))
    monkeypatch.setattr("kong.db.database.init", init)
    s = kong.get_instance()
    assert s is not None
    init.assert_called_once()
    assert s.cwd == kong.model.Folder.get_root()


def test_ls(tree, state, sample_jobs):
    root = tree

    folders, jobs = state.ls(".")
    assert all(a == b for a, b in zip(folders, root.children))
    assert all(a == b for a, b in zip(jobs, root.jobs))

    with pytest.raises(pw.DoesNotExist):
        state.ls("/nope")

    f2 = Folder.find_by_path(state.cwd, "/f2")
    state.cwd = f2
    folders, jobs = state.ls(".")
    assert all(a == b for a, b in zip(folders, f2.children))
    assert all(a == b for a, b in zip(jobs, f2.jobs))


@skip_lxplus
def test_ls_refresh(tree, state, sample_jobs):
    _, jobs = state.ls(".")
    for job in sample_jobs:
        assert job.status == Job.Status.CREATED

    for job in sample_jobs:
        job.submit()

    time.sleep(0.2)

    # without refresh
    _, jobs = state.ls(".")
    for job in jobs:
        # status is the same
        assert job.status == Job.Status.SUBMITTED

    # with refresh
    _, jobs = state.ls(".", refresh=True)
    for job in jobs:
        # status is the changes
        assert job.status == Job.Status.COMPLETED

    for job in sample_jobs:
        job.reload()

    n_completed = len([j for j in sample_jobs if j.status == Job.Status.COMPLETED])
    assert n_completed == len(state.cwd.jobs)

    _, jobs = state.ls(".", refresh=True, recursive=True)
    assert len(jobs) == len(sample_jobs)
    for job in sample_jobs:
        job.reload()
    assert all(j.status == Job.Status.COMPLETED for j in sample_jobs)


class ValidDriver(kong.drivers.local_driver.LocalDriver):
    def __init__(self, config: Config):
        pass

    def create_job(self, folder: Folder, command: str) -> Job:
        batch_job_id = str(uuid.uuid1())
        job: Job = Job.create(
            folder=folder,
            batch_job_id=batch_job_id,
            command=command,
            driver=self.__class__,
            cores=1,
        )

        return job


@skip_lxplus
def test_refresh_jobs(state, sample_jobs, monkeypatch):
    _, jobs = state.ls(".")
    for job in jobs:
        assert job.status == Job.Status.CREATED

    for job in sample_jobs:
        job.submit()

    time.sleep(0.2)

    # with refresh
    jobs = state.refresh_jobs(sample_jobs)
    for job in jobs:
        # status is the changes
        assert job.status == Job.Status.COMPLETED

    driver = ValidDriver(state.config)

    # create more jobs, some with local, some with "valid" driver
    jobs = [
        state.default_driver.create_job(command="sleep 0.1", folder=state.cwd),
        driver.create_job(command="sleep 0.1", folder=state.cwd),
        state.default_driver.create_job(command="sleep 0.1", folder=state.cwd),
        state.default_driver.create_job(command="sleep 0.1", folder=state.cwd),
        state.default_driver.create_job(command="sleep 0.1", folder=state.cwd),
    ]

    for job in jobs[:1] + jobs[2:]:
        job.submit()

    time.sleep(0.2)

    state.refresh_jobs(jobs)

    assert all(
        [
            a == b
            for a, b in zip(
                [j.status for j in jobs],
                [
                    Job.Status.COMPLETED,
                    Job.Status.CREATED,
                    Job.Status.COMPLETED,
                    Job.Status.COMPLETED,
                    Job.Status.COMPLETED,
                ],
            )
        ]
    )


def test_cd(state):
    root = Folder.get_root()
    assert state.cwd == root

    with pytest.raises(pw.DoesNotExist):
        state.cd("nope")
    assert state.cwd == root

    nope = root.add_folder("nope")

    state.cd("nope")
    assert state.cwd == nope

    state.cd("")
    assert state.cwd == root

    with pytest.raises(pw.DoesNotExist):
        state.cd("..")
    assert state.cwd == root

    with pytest.raises(pw.DoesNotExist):
        state.cd("../nope")
    assert state.cwd == root

    more = root.add_folder("more")
    another = nope.add_folder("another")

    state.cd("/nope")
    assert state.cwd == nope

    state.cd("/nope/another")
    assert state.cwd == another

    with pytest.raises(pw.DoesNotExist):
        state.cd("/../")
    assert state.cwd == another

    state.cd("..")
    assert state.cwd == nope

    state.cd("/more")
    assert state.cwd == more

    state.cd(root)
    assert state.cwd == root

    state.cd(more)
    assert state.cwd == more


def test_cd_invalid(state):
    with pytest.raises(TypeError):
        state.cd(42)


def test_mv_folder(state, db):
    root = Folder.get_root()

    f1, f2, f3, f4, f5 = [root.add_folder(n) for n in ("f1", "f2", "f3", "f4", "f5")]

    assert len(root.children) == 5

    # actual move
    state.mv("f1", f2)
    assert len(root.children) == 4
    assert len(f2.children) == 1 and f2.children[0] == f1
    f1.reload()
    assert f1.parent == f2

    # rename f3 -> f3x
    state.mv(f3, "f3x")
    assert len(root.children) == 4
    assert f3.name == "f3x"

    # another move
    state.mv(f3, "f4")
    assert len(f4.children) == 1 and f4.children[0] == f3
    assert f3.parent == f4
    assert f3.name == "f3x"

    # move rename at the same time
    state.cd("f2")
    state.mv(f5, "../f4/f5x")
    assert len(f4.children) == 2
    assert f5.name == "f5x"
    assert f5.parent == f4

    # try move to nonexistant
    state.cd("/")
    with pytest.raises(ValueError):
        state.mv(f1, "/nope/blub")

    # try to move nonexistant
    with pytest.raises(DoesNotExist):
        state.mv("../nope", f1)


def test_mv_job(state, db):
    root = Folder.get_root()

    f1, f2 = [root.add_folder(n) for n in ("f1", "f2")]

    assert len(root.children) == 2

    j1, j2, j3, j4, j5 = [state.create_job(command="sleep 1") for _ in range(5)]
    assert len(root.jobs) == 5

    state.mv(j1, "f1")
    j1.reload()
    assert j1.folder == f1
    assert len(f1.jobs) == 1
    assert len(root.jobs) == 4

    state.mv(str(j2.job_id), f2)
    j2.reload()
    assert j2.folder == f2
    assert len(f2.jobs) == 1
    assert len(root.jobs) == 3

    state.cd(f2)
    state.mv(j3, ".")
    j3.reload()
    assert j3.folder == f2

    state.mv(j2, "..")
    j2.reload()
    assert j2.folder == root

    state.mv(f"../{j4.job_id}", "../f1")
    j4.reload()
    assert j4.folder == f1

    state.cd(root)

    # renaming does not work
    with pytest.raises(ValueError):
        state.mv(f"{j5.job_id}", "42")


def test_mv_bulk_job(state):
    root = Folder.get_root()

    f1, f2, f3 = [root.add_folder(n) for n in ("f1", "f2", "f3")]
    assert len(root.children) == 3

    state.cwd = f1
    j1, j2, j3, j4, j5 = [state.create_job(command="sleep 1") for _ in range(5)]
    assert len(f1.jobs) == 5

    state.cwd = root

    state.mv("f1/*", f3)
    assert len(f1.jobs) == 0 and len(f3.jobs) == 5
    assert len(root.children) == 3
    for j in (j1, j2, j3, j4, j5):
        j.reload()
        assert j.folder == f3

    state.cwd = f2
    state.mv("../f3/*", ".")
    assert len(f3.jobs) == 0 and len(f2.jobs) == 5
    for j in (j1, j2, j3, j4, j5):
        j.reload()
        assert j.folder == f2


def test_mv_bulk_both(state):
    root = Folder.get_root()
    f1, f2 = [root.add_folder(n) for n in ("f1", "f2")]
    f3 = f1.add_folder("f3")

    with state.pushd(f1):
        j1 = state.create_job(command="sleep 1")

    state.mv("f1/*", "f2")
    j1.reload()
    assert j1.folder == f2
    f3.reload()
    assert f3.parent == f2

    f3.parent = root
    f3.save()
    j1.folder = root
    j1.save()

    # attempt to move all in root to f2. this will fail for f2, but only for f2
    state.mv("*", "f2")
    for o in (f1, f2, f3, j1):
        o.reload()
    assert f1.parent == f2
    assert f2.parent == root
    assert f3.parent == f2
    assert j1.folder == f2


def test_mv_invalid(state):
    root = Folder.get_root()
    f1, f2 = [root.add_folder(n) for n in ("f1", "f2")]
    f3 = f1.add_folder("f3")

    job = state.create_job(command="sleep 1")

    with pytest.raises(TypeError):
        state._mv_folder(f1, 42)

    with pytest.raises(TypeError):
        state._mv_folders([f1, f2], 42)

    with pytest.raises(TypeError):
        state.mv(f1, 42)

    with pytest.raises(TypeError):
        state.mv(42, f1)

    with pytest.raises(TypeError):
        state._mv_jobs(job, 42)


def test_cwd_context_manager(state):
    root = Folder.get_root()
    f1 = root.add_folder("f1")
    assert state.cwd == root

    with state.pushd(f1):
        assert state.cwd == f1
    assert state.cwd == root

    with state.pushd("f1"):
        assert state.cwd == f1
    assert state.cwd == root

    with pytest.raises(TypeError):
        with state.pushd(42):
            pass


def test_mv_bulk_folder(state):
    root = Folder.get_root()

    r1, r2 = [root.add_folder(n) for n in ("r1", "r2")]

    folders = [r1.add_folder(f"f{n}") for n in range(5)]
    assert len(r1.children) == len(folders)
    assert len(r2.children) == 0

    state.mv("r1/*", "r2")
    assert len(r1.children) == 0
    assert len(r2.children) == len(folders)
    for f in folders:
        f.reload()
        assert f.parent == r2

    state.cwd = r1
    state.mv("../r2/*", ".")
    assert len(r1.children) == len(folders)
    assert len(r2.children) == 0
    for f in folders:
        f.reload()
        assert f.parent == r1


def test_get_folders(state):
    root = Folder.get_root()

    folders = [root.add_folder(f"f{n}") for n in range(10)]

    globbed = state.get_folders("*")

    assert len(globbed) == len(folders)
    for a, b in zip(globbed, folders):
        assert a == b

    with pytest.raises(ValueError):
        state.get_folders("nope/*")


def test_get_folders_pattern(state):
    root = Folder.get_root()

    folders_alpha = [root.add_folder(f"alpha_{n}") for n in range(10)]
    folders_beta = [root.add_folder(f"beta_{n}") for n in range(10)]

    globbed_alpha = state.get_folders("alpha_*")
    assert len(globbed_alpha) == len(folders_alpha)
    assert all(a == b for a, b in zip(globbed_alpha, folders_alpha))

    globbed_beta = state.get_folders("beta_*")
    assert len(globbed_beta) == len(folders_beta)
    assert all(a == b for a, b in zip(globbed_beta, folders_beta))


def test_mkdir(state, db):
    root = Folder.get_root()
    sub = root.add_folder("sub")
    for cwd in [root, sub]:
        state.cwd = cwd

        assert cwd.subfolder("alpha") is None
        state.mkdir("alpha")
        alpha = cwd.subfolder("alpha")
        assert alpha is not None

        # one down
        assert alpha.subfolder("beta") is None
        state.mkdir("alpha/beta")
        beta = alpha.subfolder("beta")
        assert beta is not None

        # cannot create outside of root
        if cwd == root:
            with pytest.raises(kong.state.CannotCreateError):
                state.mkdir("../nope")
        else:
            state.mkdir("../nope")
            assert root.subfolder("nope") is not None

        # cannot create again
        with pytest.raises(kong.state.CannotCreateError):
            state.mkdir("alpha")

        # cannot create in nonexistant
        with pytest.raises(kong.state.CannotCreateError):
            state.mkdir("omega/game")

        state.cwd = beta
        assert cwd.subfolder("gamma") is None
        state.mkdir("../../gamma")
        gamma = cwd.subfolder("gamma")
        assert gamma is not None


def test_mkdir_existing(state, db):
    root = Folder.get_root()
    f1 = root.add_folder("f1")
    assert len(root.children) == 1

    with pytest.raises(CannotCreateError):
        state.mkdir("f1")
    assert len(root.children) == 1
    state.mkdir("f1", exist_ok=True)  # silently noop
    assert len(root.children) == 1


def test_mkdir_create_parent(state):
    root = Folder.get_root()

    with pytest.raises(CannotCreateError):
        state.mkdir("/a1/b2/c3/d4")
    assert Folder.find_by_path(state.cwd, "/a1/b2/c3/d4") is None

    state.mkdir("/a1/b2/c3/d4", create_parent=True)
    assert Folder.find_by_path(state.cwd, "/a1") is not None
    assert Folder.find_by_path(state.cwd, "/a1/b2") is not None
    assert Folder.find_by_path(state.cwd, "/a1/b2/c3") is not None
    assert Folder.find_by_path(state.cwd, "/a1/b2/c3/d4") is not None

    state.mkdir("/a1/b2/c3/d5", create_parent=True)

    state.mkdir("beta", create_parent=True)
    assert Folder.find_by_path(state.cwd, "/beta") is not None

    state.mkdir("basic", create_parent=True)
    assert Folder.find_by_path(state.cwd, "/basic") is not None

    state.mkdir("/basic2", create_parent=True)
    assert Folder.find_by_path(state.cwd, "/basic2") is not None

def test_rm_folder(state, db):
    root = Folder.get_root()

    with pytest.raises(DoesNotExist):
        state.rm("../nope")

    with pytest.raises(kong.state.CannotRemoveRoot):
        state.rm("/")

    root.add_folder("alpha")

    with pytest.raises(kong.state.CannotRemoveIsFolder):
        confirm = Mock(return_value=False)
        state.rm("alpha", confirm=confirm)
    assert confirm.call_count == 0

    assert root.subfolder("alpha") is not None

    confirm = Mock(return_value=False)
    state.rm("alpha", confirm=confirm, recursive=True)
    confirm.assert_called_once()

    assert root.subfolder("alpha") is not None

    with pytest.raises(kong.state.CannotRemoveIsFolder):
        confirm = Mock(return_value=True)
        state.rm("alpha", confirm=confirm)
    assert confirm.call_count == 0

    assert root.subfolder("alpha") is not None

    confirm = Mock(return_value=True)
    state.rm("alpha", confirm=confirm, recursive=True)
    confirm.assert_called_once()
    assert root.subfolder("alpha") is None

    # works further down
    beta = root.add_folder("beta")
    gamma = beta.add_folder("gamma")
    assert beta.subfolder("gamma") is not None
    state.rm("/beta/gamma", recursive=True)
    assert beta.subfolder("gamma") is None

    # should also work with instance
    assert root.subfolder("beta") is not None
    state.rm(beta, recursive=True)
    assert root.subfolder("beta") is None


def test_rm_job(state, db):
    root = Folder.get_root()
    j1 = state.create_job(command="sleep 1")
    assert len(root.jobs) == 1 and root.jobs[0] == j1
    assert Job.get_or_none(job_id=j1.job_id) is not None
    state.rm(str(j1.job_id))  # confirm default is True
    assert len(root.jobs) == 0
    assert Job.get_or_none(job_id=j1.job_id) is None

    # should also work with instance
    j2 = state.create_job(command="sleep 1")
    assert len(root.jobs) == 1 and root.jobs[0] == j2
    assert Job.get_or_none(job_id=j2.job_id) is not None
    confirm = Mock(return_value=False)
    state.rm(j2, confirm=confirm)
    confirm.assert_called_once()
    # no change
    assert len(root.jobs) == 1 and root.jobs[0] == j2
    assert Job.get_or_none(job_id=j2.job_id) is not None

    # now change
    confirm = Mock(return_value=True)
    state.rm(j2, confirm=confirm)
    confirm.assert_called_once()
    assert len(root.jobs) == 0
    assert Job.get_or_none(job_id=j2.job_id) is None

    # works in other cwd too
    alpha = root.add_folder("alpha")
    j3 = state.default_driver.create_job(command="sleep 1", folder=alpha)
    assert len(alpha.jobs) == 1 and alpha.jobs[0] == j3
    assert Job.get_or_none(job_id=j3.job_id) is not None
    state.rm(str(j3.job_id))
    assert len(alpha.jobs) == 0
    assert Job.get_or_none(job_id=j3.job_id) is None


def test_rm_recursive(state, sample_jobs):
    all_jobs = Job.select().execute()
    assert len(all_jobs) == len(sample_jobs)

    root = Folder.get_root()
    for f in root.children:
        confirm = Mock()
        with pytest.raises(CannotRemoveIsFolder):
            state.rm(f, confirm=confirm)
        assert confirm.call_count == 0

        confirm = Mock(return_value=False)
        assert state.rm(f, recursive=True, confirm=confirm) == False
        assert confirm.call_count == 1

        confirm = Mock(return_value=True)
        assert state.rm(f, recursive=True, confirm=confirm) == True
        assert confirm.call_count == 1

    all_jobs = Job.select().execute()
    assert len(all_jobs) == len(root.jobs)  # only root jobs remain


def test_rm_invalid(state):
    with pytest.raises(TypeError):
        state.rm(4.5)


def test_get_driver(state, db):
    driver = state.default_driver
    assert isinstance(driver, LocalDriver)


def test_create_job(state, db):
    root = Folder.get_root()
    j1 = state.create_job(command="sleep 1")
    assert j1.folder == root
    assert len(root.jobs) == 1 and root.jobs[0] == j1

    f2 = root.add_folder("f2")
    state.cd("f2")
    j2 = state.create_job(command="sleep 1")
    assert j2.folder == f2
    assert len(f2.jobs) == 1 and f2.jobs[0] == j2

    # no explicit folder
    with pytest.raises(ValueError):
        state.create_job(command="a", folder="blub")

    # no explicit driver
    with pytest.raises(ValueError):
        state.create_job(command="a", driver="blub")


def test_get_jobs(state):
    root = Folder.get_root()
    j1 = state.create_job(command="sleep 1")

    f2 = root.add_folder("f2")
    state.cd("f2")
    j2 = state.create_job(command="sleep 1")
    j3 = state.create_job(command="sleep 1")

    assert state.get_jobs(f"{j1.job_id}")[0] == j1
    assert state.get_jobs(f"f2/{j2.job_id}")[0] == j2
    with pytest.raises(DoesNotExist):
        state.get_jobs("42")

    assert all(a == b for a, b in zip(state.get_jobs("../*"), [j1]))
    assert all(a == b for a, b in zip(state.get_jobs("*"), [j2, j3]))
    state.cwd = root
    assert all(a == b for a, b in zip(state.get_jobs("f2/*"), [j2, j3]))

    # get by id that does not exist
    with pytest.raises(DoesNotExist):
        state.get_jobs(42)

    # get path/id that does not exist
    with pytest.raises(DoesNotExist):
        state.get_jobs("/42")

    # get single job instance (this is a bit redundant)
    assert state.get_jobs(j1) == [j1]

    with pytest.raises(TypeError):
        state.get_jobs(4.2)


def test_get_folders_jobs_pattern(state):
    root = Folder.get_root()

    folders_alpha = [root.add_folder(f"alpha_{n}") for n in range(10)]
    folders_beta = [root.add_folder(f"beta_{n}") for n in range(13)]

    jobs_alpha = []
    for f in folders_alpha:
        with state.pushd(f):
            jobs_alpha.append(state.create_job(command="sleep 1"))

    jobs_beta = []
    for f in folders_beta:
        with state.pushd(f):
            jobs_beta.append(state.create_job(command="sleep 1"))

    globbed_alpha = state.get_jobs("alpha_*/*")
    assert len(globbed_alpha) == len(jobs_alpha)
    assert all(a == b for a, b in zip(globbed_alpha, jobs_alpha))

    globbed_beta = state.get_jobs("beta_*/*")
    assert len(globbed_beta) == len(jobs_beta)
    jobid = lambda j: j.job_id
    assert all(
        a == b
        for a, b in zip(sorted(globbed_beta, key=jobid), sorted(jobs_beta, key=jobid))
    )


def test_get_jobs_range(state):
    root = Folder.get_root()

    f1, f2, f3, f4, f5 = [state.create_job(command="sleep 1") for _ in range(5)]

    jobs = state.get_jobs("2..4")
    assert len(jobs) == 3
    assert set([f2, f3, f4]) == set(jobs)

    jobs = state.get_jobs("1..2")
    assert len(jobs) == 2
    assert set([f1, f2]) == set(jobs)

    jobs = state.get_jobs("2..5")
    assert len(jobs) == 4
    assert set([f2, f3, f4, f5]) == set(jobs)

    with pytest.raises(ValueError):
        state.get_jobs("4..2")


def test_submit_job(state, db):
    root = Folder.get_root()

    j1 = state.create_job(command="sleep 1")
    assert j1.status == Job.Status.CREATED

    confirm = Mock(return_value=False)
    state.submit_job(j1.job_id, confirm=confirm)
    assert confirm.call_count == 1
    assert j1.status == Job.Status.CREATED

    confirm = Mock(return_value=True)
    state.submit_job(j1.job_id, confirm=confirm)
    assert confirm.call_count == 1

    j1.reload()
    assert j1.status == Job.Status.SUBMITTED

    root.add_folder("f1")
    state.cd("f1")
    j3 = state.create_job(command="sleep 1")
    j4 = state.create_job(command="sleep 1")
    assert j3.status == Job.Status.CREATED
    state.cd("..")
    assert state.cwd == root
    state.submit_job(j3.job_id)
    j3.reload()
    assert j3.status == Job.Status.SUBMITTED

    assert j4.status == Job.Status.CREATED
    assert state.cwd == root
    state.submit_job(f"f1/{j4.job_id}")
    j4.reload()
    assert j4.status == Job.Status.SUBMITTED

    with pytest.raises(DoesNotExist):
        state.submit_job("4242")


def test_submit_job_non_recursive(state, sample_jobs):
    root = Folder.get_root()

    assert all(j.status == Job.Status.CREATED for j in sample_jobs)
    state.submit_job("/*")
    assert all(j.status == Job.Status.SUBMITTED for j in root.jobs)

    for j in sample_jobs:
        j.reload()

    submitted_jobs = [j for j in sample_jobs if j.status == Job.Status.SUBMITTED]
    assert all(a == b for a, b in zip(submitted_jobs, root.jobs))


def test_submit_job_recursive(state, sample_jobs):
    root = Folder.get_root()

    assert all(j.status == Job.Status.CREATED for j in sample_jobs)
    state.submit_job("/", recursive=True)

    for j in sample_jobs:
        j.reload()

    assert all(j.status == Job.Status.SUBMITTED for j in sample_jobs)


def test_kill_job(state):
    root = Folder.get_root()
    j1 = state.create_job(command="sleep 1")
    j1.submit()
    assert j1.status == Job.Status.SUBMITTED
    time.sleep(0.1)
    assert j1.get_status() == Job.Status.RUNNING

    confirm = Mock(return_value=False)
    state.kill_job(j1.job_id, confirm=confirm)
    assert confirm.call_count == 1
    assert j1.get_status() == Job.Status.RUNNING

    confirm = Mock(return_value=True)
    state.kill_job(j1.job_id, confirm=confirm)
    assert confirm.call_count == 1
    assert j1.get_status() == Job.Status.FAILED


def test_kill_job_non_recursive(state, sample_jobs):
    root = Folder.get_root()

    assert all(j.status == Job.Status.CREATED for j in sample_jobs)
    state.kill_job("/*")
    assert all(j.status == Job.Status.FAILED for j in root.jobs)

    for j in sample_jobs:
        j.reload()

    failed_jobs = [j for j in sample_jobs if j.status == Job.Status.FAILED]
    for a, b in zip(failed_jobs, root.jobs):
        assert a == b


def test_kill_job_recursive(state, sample_jobs):
    root = Folder.get_root()

    assert all(j.status == Job.Status.CREATED for j in sample_jobs)
    state.kill_job("/", recursive=True)

    for j in sample_jobs:
        j.reload()

    assert all(j.status == Job.Status.FAILED for j in sample_jobs)


def test_job_resubmit(state):
    root = Folder.get_root()
    j1 = state.create_job(command="sleep 1")
    j1.submit()
    assert j1.status == Job.Status.SUBMITTED
    time.sleep(0.1)
    assert j1.get_status() == Job.Status.RUNNING
    state.kill_job(j1.job_id)
    assert j1.get_status() == Job.Status.FAILED

    confirm = Mock(return_value=False)
    state.resubmit_job(str(j1.job_id), confirm=confirm)
    assert confirm.call_count == 1
    assert j1.get_status() == Job.Status.FAILED

    confirm = Mock(return_value=True)
    state.resubmit_job(str(j1.job_id), confirm=confirm)
    assert confirm.call_count == 1

    j1.reload()
    assert j1.status == Job.Status.SUBMITTED


def test_job_resubmit_non_recursive(state, sample_jobs):
    root = Folder.get_root()
    for job in sample_jobs:
        job.status = Job.Status.FAILED
        job.save()

    state.resubmit_job("/*")
    assert all(j.status == Job.Status.SUBMITTED for j in root.jobs)

    for j in sample_jobs:
        j.reload()

    submitted_jobs = [j for j in sample_jobs if j.status == Job.Status.SUBMITTED]
    assert all(a == b for a, b in zip(submitted_jobs, root.jobs))


def test_job_resubmit_recursive(state, sample_jobs):
    root = Folder.get_root()
    for job in sample_jobs:
        job.status = Job.Status.FAILED
        job.save()

    state.resubmit_job("/", recursive=True)

    for j in sample_jobs:
        j.reload()

    assert all(j.status == Job.Status.SUBMITTED for j in sample_jobs)


@skip_lxplus
def test_job_resubmit_failed_only(state):
    root = Folder.get_root()
    j1 = state.create_job(command="sleep 0.1")
    j2 = state.create_job(command="sleep 0.1")
    j3 = state.create_job(command="exit 1")

    for j in [j1, j2, j3]:
        j.submit()

    state.default_driver.wait([j1, j2, j3])

    assert j1.get_status() == Job.Status.COMPLETED
    assert j2.get_status() == Job.Status.COMPLETED
    assert j3.get_status() == Job.Status.FAILED

    state.resubmit_job("/", recursive=True, failed_only=True)
    j3.reload()
    assert j1.get_status() == Job.Status.COMPLETED
    assert j2.get_status() == Job.Status.COMPLETED
    assert j3.status == Job.Status.SUBMITTED


def test_wait(state, monkeypatch):
    root = Folder.get_root()
    j1 = state.create_job(command="sleep 0.1")
    j2 = state.create_job(command="sleep 0.1")
    j3 = state.create_job(command="exit 1")

    jobs = [[j1, j2, j3]]

    driver = Mock()
    driver.wait = Mock()
    driver.wait.return_value = iter(jobs)
    driver.bulk_sync_status = Mock(return_value=jobs[0])

    factory = Mock(return_value=driver)
    monkeypatch.setattr("kong.drivers.local_driver.LocalDriver", factory)

    monkeypatch.setattr("kong.state.Spinner", MagicMock())

    nm = Mock()
    nm.notify = Mock()
    monkeypatch.setattr(state.config, "notifications", nm)

    state.wait("*", notify=False)
    assert nm.notify.call_count == 0
    assert driver.wait.call_count == 1

    exhaust(state.wait("*", notify=True, progress=True))
    assert nm.notify.call_count == 1
    assert driver.wait.call_count == 2


def test_wait_timeout(state, monkeypatch):
    root = Folder.get_root()
    j1 = state.create_job(command="sleep 0.1")
    j2 = state.create_job(command="sleep 0.1")
    j3 = state.create_job(command="exit 1")

    jobs = [[j1, j2, j3]]

    driver = Mock()
    driver.wait = Mock()
    driver.wait.side_effect = TimeoutError

    factory = Mock(return_value=driver)
    monkeypatch.setattr("kong.drivers.local_driver.LocalDriver", factory)

    monkeypatch.setattr("kong.state.Spinner", MagicMock())

    with monkeypatch.context() as m:
        nm = MagicMock()
        m.setattr(state.config, "notifications", nm)
        state.wait("*", notify=False)
        assert nm.notify.call_count == 0
        state.wait("*", notify=True)
        assert nm.notify.call_count == 1



import datetime
import multiprocessing
import tempfile
import os
from contextlib import contextmanager
from subprocess import Popen
from typing import (
    Any,
    Optional,
    IO,
    Union,
    List,
    Iterable,
    Dict,
    Collection,
    Iterator,
    cast,
    Sequence,
)
import uuid

import psutil

from ..util import rmtree
from ..db import database
from ..model import Folder, Job
from ..logger import logger
from . import InvalidJobStatus
from .driver_base import DriverBase, checked_job

jobscript_tpl = """
#!/usr/bin/env bash

exit_status_file={exit_status_file}
stdout={stdout}
stderr={stderr}

sig_handler() {{
    exit_status=$?
    echo $exit_status > $exit_status_file
}}
trap sig_handler INT HUP TERM QUIT

export KONG_JOB_ID={internal_job_id}
export KONG_JOB_OUTPUT_DIR={output_dir}
export KONG_JOB_LOG_DIR={log_dir}
export KONG_JOB_NPROC={nproc}
export KONG_JOB_SCRATCHDIR={scratch_dir}

touch $stdout
touch $stderr

({command}) >> $stdout 2>> $stderr
echo $? > {exit_status_file}

""".strip()


class LocalDriver(DriverBase):
    def create_job(
        self, folder: Folder, command: str, cores: int = 1, *args: Any, **kwargs: Any
    ) -> Job:
        assert len(args) == 0 and len(kwargs) == 0, "No extra arguments allowed"
        batch_job_id = str(uuid.uuid1())

        job: Job = Job.create(
            folder=folder,
            batch_job_id=batch_job_id,
            command=command,
            driver=self.__class__,
            cores=cores,
        )

        # in job dir, create output dir
        output_dir = self.make_output_path(job)
        os.makedirs(output_dir, exist_ok=True)

        log_dir = self.make_log_path(job)
        os.makedirs(log_dir, exist_ok=True)

        stdout = os.path.abspath(os.path.join(log_dir, "stdout.txt"))
        stderr = os.path.abspath(os.path.join(log_dir, "stderr.txt"))
        exit_status_file = os.path.abspath(os.path.join(log_dir, "exit_status.txt"))

        scriptpath = os.path.join(log_dir, "jobscript.sh")

        scratch_dir = tempfile.mkdtemp(prefix=f"kong_job_{job.job_id}")

        job.data = dict(
            stdout=stdout,
            stderr=stderr,
            exit_status_file=exit_status_file,
            jobscript=scriptpath,
            output_dir=output_dir,
            log_dir=log_dir,
            scratch_dir=scratch_dir,
        )
        job.save()

        values = dict(
            command=command,
            stdout=stdout,
            stderr=stderr,
            internal_job_id=job.job_id,
            output_dir=output_dir,
            log_dir=log_dir,
            exit_status_file=exit_status_file,
            nproc=cores,
            scratch_dir=scratch_dir,
        )
        logger.debug("Creating job with values: %s", str(values))

        jobscript = jobscript_tpl.format(**values)

        with open(scriptpath, "w") as fh:
            fh.write(jobscript)

        job._driver_instance = self
        return job

    def bulk_create_jobs(self, jobs: Iterable[Dict[str, Any]]) -> List["Job"]:
        # right now, implemented as loop, potential to optimize
        return [self.create_job(**kwargs) for kwargs in jobs]

    def cleanup(self, job: Job) -> Job:
        if job.status not in (
            Job.Status.CREATED,
            Job.Status.FAILED,
            Job.Status.COMPLETED,
            Job.Status.UNKNOWN,
        ):
            raise InvalidJobStatus(
                f"Cannot clean up job {job} in {job.status}, please kill first"
            )

        logger.debug("Removing job output directory for job %s", job)

        for name in ["log_dir", "output_dir", "scratch_dir"]:
            path = job.data[name]
            if os.path.exists(path):
                rmtree(path)
        return job

    def bulk_cleanup(self, jobs: Sequence["Job"]) -> Sequence["Job"]:
        for job in jobs:
            self.cleanup(job)
        return jobs

    def remove(self, job: Job) -> None:
        logger.debug("Removing job %s", job)
        self.cleanup(job)
        job.delete_instance()

    def bulk_remove(self, jobs: Sequence["Job"]) -> None:
        for job in jobs:
            self.remove(job)

    @checked_job
    def sync_status(self, job: Job, save: bool = True) -> Job:
        if job.status not in (Job.Status.RUNNING, Job.Status.SUBMITTED):
            logger.debug(
                "Job %s is neither RUNNING nor SUBMITTED (%s), so no status changes without intervention",
                job,
                job.status,
            )
            return job

        logger.debug("Job %s in status %s, checking for updates", job, job.status)
        exit_status_file = job.data["exit_status_file"]
        pid = job.data["pid"]

        def check_exit_code() -> None:
            if not os.path.exists(exit_status_file):
                logger.debug(
                    "Job %s appears to have exited, but exit status file is not present"
                )
                job.status = Job.Status.UNKNOWN
            else:
                with open(exit_status_file, "r") as fh:
                    exit_code = int(fh.read().strip())
                job.data["exit_code"] = exit_code
                if exit_code == 0:
                    logger.debug("Job %s succeeded", job)
                    job.status = Job.Status.COMPLETED
                else:
                    logger.debug("Job %s failed", job)
                    job.status = Job.Status.FAILED

        # check if it is still running
        try:
            proc = psutil.Process(pid)
            if proc.is_running():
                # is running, but is it zombie waiting to be reaped?
                if proc.status() == psutil.STATUS_ZOMBIE:  # pragma: no cover
                    logger.debug(
                        "Job %s with pid %s is running but zombie, reaping", job, pid
                    )
                    proc.wait()  # reaping
                    logger.debug("Reaped pid %d", pid)
                    check_exit_code()
                else:
                    job.status = Job.Status.RUNNING
            else:
                logger.debug("Job %s is not running, exit code should be set", job)
                check_exit_code()
            if save:
                job.save()

        except psutil.NoSuchProcess:
            logger.debug("Job %s with pid %d doesn't exist, check exit code", job, pid)
            check_exit_code()
            if save:
                job.save()
        return job

    def bulk_sync_status(self, jobs: Sequence[Job]) -> Sequence[Job]:
        # simply implemented as loop over single sync status for local driver
        now = datetime.datetime.now()

        def sync() -> Iterable[Job]:
            for job in jobs:
                self.sync_status(job, save=False)
                job.updated_at = now
                yield job

        with database.atomic():
            Job.bulk_update(
                sync(), fields=[Job.status, Job.updated_at], batch_size=self.batch_size
            )

        return jobs

    @checked_job
    def kill(self, job: Job, save: bool = True) -> None:
        self.sync_status(job)
        if job.status == Job.Status.CREATED:
            logger.debug("Job %s in %s, simply setting to failed", job, job.status)
            job.status = Job.Status.FAILED
        elif job.status in (Job.Status.RUNNING, Job.Status.SUBMITTED):
            logger.debug(
                "Job %s in %s, killing pid %d", job, job.status, job.data["pid"]
            )
            proc = psutil.Process(job.data["pid"])
            proc.kill()
            proc.wait()
            job.status = Job.Status.FAILED
        else:
            logger.debug("Job %s in %s, do nothing")
        if save:
            job.save()

    def bulk_kill(self, jobs: Sequence["Job"]) -> Sequence[Job]:
        now = datetime.datetime.now()

        def k() -> Iterable[Job]:
            for job in jobs:
                self.kill(job, save=False)
                job.updated_at = now
                yield job

        with database.atomic():
            Job.bulk_update(
                k(), fields=[Job.status, Job.updated_at], batch_size=self.batch_size
            )

        return jobs

    def bulk_submit(self, jobs: Iterable["Job"]) -> None:
        now = datetime.datetime.now()

        def sub() -> Iterable[Job]:
            for job in jobs:
                self.submit(job, save=False)
                job.updated_at = now
                yield job

        with database.atomic():
            Job.bulk_update(
                sub(),
                fields=[Job.status, Job.data, Job.updated_at],
                batch_size=self.batch_size,
            )

    @classmethod
    def spawn_child(
        cls, cmd: List[str], pid: multiprocessing.Value
    ) -> None:  # pragma: no cover
        os.setsid()  # deamonize so we detach from handlers, avoid zombie state
        proc = Popen(cmd, stdin=None, stdout=None, stderr=None, close_fds=True)
        pid.value = proc.pid

    @checked_job
    def submit(self, job: Job, save: bool = True) -> None:
        self.sync_status(job)
        if job.status > Job.Status.CREATED:
            raise InvalidJobStatus(f"Cannot submit job in state {job.status}")

        cmd = ["/usr/bin/env", "bash", job.data["jobscript"]]
        logger.debug("About to submit job with command: %s", str(cmd))

        pid = multiprocessing.Value("i", 0)

        logger.debug("Double fork child: spawning process")
        p = multiprocessing.Process(target=LocalDriver.spawn_child, args=(cmd, pid))
        p.start()
        p.join()
        logger.debug("Double fork child: terminating")
        assert pid.value != 0, "Got invalid pid 0"
        logger.debug("Got pid: %d", pid.value)
        job.data["pid"] = pid.value
        job.status = Job.Status.SUBMITTED

        if save:
            job.save()

        logger.debug("Submitted job as %s", job)

    @checked_job  # type: ignore
    @contextmanager  # type: ignore
    def stdout(self, job: Job) -> Iterator[IO[str]]:
        self.sync_status(job)
        if job.status not in (Job.Status.FAILED, Job.Status.COMPLETED):
            raise InvalidJobStatus("Cannot get stdout for job in status %s", job.status)

        with open(job.data["stdout"], "r") as fh:
            yield fh

    @checked_job  # type: ignore
    @contextmanager  # type: ignore
    def stderr(self, job: Job) -> Iterator[IO[str]]:
        self.sync_status(job)
        if job.status not in (Job.Status.FAILED, Job.Status.COMPLETED):
            raise InvalidJobStatus("Cannot get stdout for job in status %s", job.status)

        with open(job.data["stderr"], "r") as fh:
            yield fh

    @checked_job
    def _wait_single(self, job: Job, timeout: Optional[int] = None) -> Job:
        logger.debug("Wait for job %s requested", job)
        self.sync_status(job)
        if job.status not in (Job.Status.SUBMITTED, Job.Status.RUNNING):
            logger.info(
                "Job %s is in status %s, neither SUBMITTED nor RUNNING, wait will not complete, returning now",
                job,
                job.status,
            )
            return job

        try:
            proc = psutil.Process(pid=job.data["pid"])
            try:
                proc.wait(timeout=timeout)
            except psutil.TimeoutExpired as e:
                raise TimeoutError(str(e))
        except psutil.NoSuchProcess:
            logger.debug("Wait encountered non-existing pid, we're done")

        return cast(Job, self.sync_status(job))

    def wait_gen(
        self,
        jobs: Union[Job, List[Job]],
        poll_interval: Optional[int] = None,
        timeout: Optional[int] = None,
    ) -> Iterable[List[Job]]:
        if not isinstance(jobs, list):
            jobs = [jobs]

        logger.debug("Waiting for %s jobs", len(jobs))
        for job in jobs:
            self._wait_single(job, timeout=timeout)

        yield list(self.bulk_sync_status(jobs))

    @checked_job
    def resubmit(self, job: Job) -> Job:
        self.sync_status(job)
        if job.status not in (
            Job.Status.COMPLETED,
            Job.Status.FAILED,
            Job.Status.UNKNOWN,
        ):
            logger.error("Will not resubmit job %s in status %s", job, job.status)
            raise InvalidJobStatus(
                f"Will not resubmit job {job} in status {job.status}"
            )

        self.kill(job)

        # need to make sure the output artifacts are gone, since we're reusing the same job dir
        for name in ["exit_status_file", "stdout", "stderr"]:
            path = job.data[name]
            if os.path.exists(path):
                logger.debug("Removing %s", path)
                os.remove(path)
            assert not os.path.exists(path)

        for d in ["scratch_dir", "output_dir"]:
            path = job.data[d]
            if os.path.exists(path):
                logger.debug("Removing %s", path)
                rmtree(path)
                os.makedirs(path)

        job.status = Job.Status.CREATED
        job.save()
        self.submit(job)
        return job

    def bulk_resubmit(
        self, jobs: Collection["Job"], do_submit: bool = True
    ) -> Iterable["Job"]:
        for job in jobs:
            self.sync_status(job)
            if job.status not in (
                Job.Status.COMPLETED,
                Job.Status.FAILED,
                Job.Status.UNKNOWN,
            ):
                logger.error("Will not resubmit job %s in status %s", job, job.status)
                raise InvalidJobStatus(
                    f"Will not resubmit job {job} in status {job.status}"
                )

            self.kill(job)

            # need to make sure the output artifacts are gone, since we're reusing the same job dir
            for name in ["exit_status_file", "stdout", "stderr"]:
                path = job.data[name]
                if os.path.exists(path):
                    logger.debug("Removing %s", path)
                    os.remove(path)
                assert not os.path.exists(path)

            for d in ["scratch_dir", "output_dir"]:
                path = job.data[d]
                if os.path.exists(path):
                    logger.debug("Removing %s", path)
                    rmtree(path)
                    os.makedirs(path)

            job.status = Job.Status.CREATED
            job.save()
            if do_submit:
                self.submit(job)
        return jobs



# flake8: noqa
from typing import Any, List, Iterable, TypeVar

import peewee as pw

from ..util import chunks
from .. import db

__all__ = ["Folder", "Job"]

T = TypeVar("T", bound="BaseModel")


class BaseModel(pw.Model):
    class Meta:
        database = db.database

    def reload(self) -> None:
        newer_self = type(self).get_by_id(self._pk)  # type: ignore
        for field_name in self._meta.fields.keys():  # type: ignore
            val = getattr(newer_self, field_name)  # type: ignore
            setattr(self, field_name, val)  # type: ignore
        self._dirty.clear()  # type: ignore

    @classmethod
    def bulk_select(
        cls, field: Any, values: List[Any], batch_size: int = 999
    ) -> Iterable[T]:
        for chunk in chunks(values, batch_size):
            yield from cls.select().where(field.in_(chunk)).execute()  # type: ignore


from .folder import Folder
from .job import Job



from typing import Optional


class NoSuchProcess(BaseException): ...
class TimeoutExpired(BaseException): ...

STATUS_ZOMBIE: int

class Process:

    def __init__(self, pid: int): ...
    def is_running(self) -> bool: ...
    def status(self) -> int: ...
    def wait(self, timeout: Optional[int] = None) -> None: ...
    def kill(self) -> None: ...



import os
import logging
from typing import Any

import click
import coloredlogs

from . import config
from . import setup
from .state import State
from .logger import logger
from .repl import Repl

import pkg_resources  # part of setuptools

version = pkg_resources.get_distribution("kong-batch").version


@click.group(invoke_without_command=True)
@click.option("--version", "show_version", is_flag=True)
@click.option("-v", "--verbose", "verbosity", count=True)
@click.pass_context
def main(ctx: Any, show_version: bool, verbosity: int) -> None:
    if verbosity == 0:
        level = logging.WARNING
        global_level = logging.WARNING
    elif verbosity <= 1:
        level = logging.INFO
        global_level = logging.INFO
    elif verbosity <= 2:
        level = logging.DEBUG
        global_level = logging.INFO
    else:
        level = logging.DEBUG
        global_level = logging.DEBUG

    coloredlogs.install(
        fmt="%(asctime)s %(levelname)s %(name)s %(filename)s:%(funcName)s %(message)s",
        level=level,
    )

    logger.setLevel(level)
    logging.getLogger().setLevel(global_level)

    if show_version:
        click.echo(f"{config.APP_NAME} version: {version}")
        return

    # check if setup was executed
    if not os.path.exists(config.CONFIG_FILE):
        logger.debug(
            "Setup was not executed yet, config file at %s does not exist",
            config.CONFIG_FILE,
        )
        try:
            setup.setup(None)
        except Exception as e:
            logger.error("Got error during setup", exc_info=True)
            raise click.ClickException(str(e))
    else:
        logger.debug("Setup executed already")
    logger.debug("Setup completed")

    inst = State.get_instance()

    ctx.obj = inst

    if ctx.invoked_subcommand is None:
        try:
            logger.debug("Entering REPL")
            Repl(ctx.obj).cmdloop()
        except Exception as e:
            logger.error("Exception caught", exc_info=True)
            raise click.ClickException(str(e))


@main.command("setup")
@click.pass_obj
def setup_command(state: State) -> None:
    setup.setup(state.config)


@main.command()
@click.pass_obj
def interactive(state: State) -> None:
    logger.debug("Launching IPython session")
    import IPython

    IPython.embed(colors="neutral")



import os
from unittest.mock import Mock, call
from unittest import mock

import yaml
import pytest

import kong.setup
import kong


def test_setup(app_env, monkeypatch):
    app_dir, config_path, tmp_path = app_env

    # first time setup
    with monkeypatch.context() as m:
        prompt = Mock(
            side_effect=[
                "kong.drivers.local_driver.LocalDriver",
                os.path.join(tmp_path, "jobs"),
                os.path.join(tmp_path, "joboutput"),
            ]
        )
        m.setattr("click.prompt", prompt)
        kong.setup.setup(None)
        prompt.assert_has_calls(
            [
                call(mock.ANY, default="kong.drivers.local_driver.LocalDriver"),
                call(mock.ANY, default=os.path.join(app_dir, "jobdir")),
                call(mock.ANY, default=os.path.join(app_dir, "joboutput")),
            ]
        )

    assert os.path.exists(config_path)

    with open(config_path) as f:
        data = yaml.safe_load(f)
    assert data["default_driver"] == "kong.drivers.local_driver.LocalDriver"
    assert data["jobdir"] == os.path.join(tmp_path, "jobs")
    assert data["joboutputdir"] == os.path.join(tmp_path, "joboutput")
    assert "history_length" in data

    # re-run setup
    cfg = kong.config.Config()
    assert cfg.data == data
    assert cfg.jobdir == os.path.join(tmp_path, "jobs")

    with monkeypatch.context() as m:
        prompt = Mock(
            side_effect=[
                "kong.drivers.local_driver.LocalDriver",
                os.path.join(tmp_path, "jobs_new"),
                os.path.join(tmp_path, "joboutput_new"),
            ]
        )
        m.setattr("click.prompt", prompt)
        kong.setup.setup(cfg)
        # should provide default from before
        prompt.assert_has_calls(
            [
                call(mock.ANY, default="kong.drivers.local_driver.LocalDriver"),
                call(mock.ANY, default=os.path.join(tmp_path, "jobs")),
                call(mock.ANY, default=os.path.join(tmp_path, "joboutput")),
            ]
        )

    with open(config_path) as f:
        data = yaml.safe_load(f)
    assert (
        data["default_driver"] == "kong.drivers.local_driver.LocalDriver"
    )  # still the same
    assert data["jobdir"] == os.path.join(tmp_path, "jobs_new")  # this has changed
    assert data["joboutputdir"] == os.path.join(
        tmp_path, "joboutput_new"
    )  # this has changed


def test_setup_invalid_driver(app_env, monkeypatch):
    app_dir, config_path, tmp_path = app_env

    with monkeypatch.context() as m:
        prompt = Mock(side_effect=["NotADriver", os.path.join(tmp_path, "jobs")])
        m.setattr("click.prompt", prompt)
        with pytest.raises(ValueError):
            kong.setup.setup(None)
        prompt.assert_has_calls(
            [call(mock.ANY, default="kong.drivers.local_driver.LocalDriver")]
        )
